import builtins

import numpy as np
import tensorflow as tf
from tensorflow.compiler.tf2xla.python.xla import dynamic_update_slice

from keras.src import tree
from keras.src.backend.common import KerasVariable
from keras.src.backend.common import global_state
from keras.src.backend.common import is_int_dtype
from keras.src.backend.common import standardize_dtype
from keras.src.backend.common.backend_utils import slice_along_axis
from keras.src.backend.common.keras_tensor import KerasTensor
from keras.src.backend.common.name_scope import name_scope as base_name_scope
from keras.src.backend.common.stateless_scope import StatelessScope
from keras.src.backend.common.stateless_scope import in_stateless_scope
from keras.src.backend.common.symbolic_scope import SymbolicScope
from keras.src.backend.tensorflow.sparse import sparse_to_dense
from keras.src.utils.naming import auto_name

SUPPORTS_SPARSE_TENSORS = True
SUPPORTS_RAGGED_TENSORS = True
SUPPORTS_COMPLEX_DTYPES = True
# https://github.com/tensorflow/tensorflow/issues/78338
IS_THREAD_SAFE = False


class Variable(
    KerasVariable,
    tf.__internal__.types.Tensor,
    tf.__internal__.tracking.Trackable,
):
    _should_act_as_resource_variable = True

    @property
    def handle(self):
        return self.value.handle

    def _initialize(self, value):
        if isinstance(value, tf.Variable):
            self._value = value
        else:
            self._value = tf.Variable(
                value,
                dtype=self._dtype,
                trainable=self.trainable,
                name=self.name,
                aggregation=self._map_aggregation(self.aggregation),
                synchronization=self._map_synchronization(self.synchronization),
            )

    def _initialize_with_initializer(self, initializer):
        self._initialize(lambda: initializer(self._shape, dtype=self._dtype))

    def _deferred_initialize(self):
        if self._value is not None:
            raise ValueError(f"Variable {self.path} is already initialized.")

        if in_stateless_scope():
            raise ValueError(
                "You are attempting to initialize a variable "
                "while in a stateless scope. This is disallowed. "
                "Make sure that all variables are initialized "
                "before you start using your layer/model objects."
            )
        with tf.init_scope():
            self._initialize_with_initializer(self._initializer)
            self._initializer = None

    def _direct_assign(self, value):
        self._value.assign(tf.cast(value, self._value.dtype))

    def _convert_to_tensor(self, value, dtype=None):
        return convert_to_tensor(value, dtype=dtype)

    def numpy(self):  # noqa: F811
        return self.value.numpy()

    @property
    def shape(self):
        return tf.TensorShape(super().shape)

    # Overload native accessor.
    def __tf_tensor__(self, dtype=None, name=None):
        return tf.convert_to_tensor(self.value, dtype=dtype, name=name)

    # Methods below are for SavedModel support
    @property
    def _shared_name(self):
        return self.value._shared_name

    def _serialize_to_tensors(self):
        try:
            return self.value._serialize_to_tensors()
        except NotImplementedError:
            return {"VARIABLE_VALUE": self.value}

    def _restore_from_tensors(self, restored_tensors):
        try:
            return self.value._restore_from_tensors(restored_tensors)
        except NotImplementedError:
            self.assign(restored_tensors["VARIABLE_VALUE"])
            return self.value

    def _copy_trackable_to_cpu(self, object_map):
        self.value._copy_trackable_to_cpu(object_map)
        object_map[self] = tf.Variable(object_map[self.value])

    def _export_to_saved_model_graph(
        self, object_map, tensor_map, options, **kwargs
    ):
        resource_list = self.value._export_to_saved_model_graph(
            object_map, tensor_map, options, **kwargs
        )
        object_map[self] = tf.Variable(object_map[self.value])
        return resource_list

    def _write_object_proto(self, proto, options):
        return self.value._write_object_proto(proto, options)

    def _map_aggregation(self, aggregation):
        mapping = {
            "none": tf.VariableAggregation.NONE,
            "sum": tf.VariableAggregation.SUM,
            "mean": tf.VariableAggregation.MEAN,
            "only_first_replica": tf.VariableAggregation.ONLY_FIRST_REPLICA,
        }
        return mapping[aggregation]

    def _map_synchronization(self, synchronization):
        mapping = {
            "none": tf.VariableSynchronization.NONE,
            "on_read": tf.VariableSynchronization.ON_READ,
            "on_write": tf.VariableSynchronization.ON_WRITE,
            "auto": tf.VariableSynchronization.AUTO,
        }
        return mapping[synchronization]


def convert_to_tensor(x, dtype=None, sparse=None, ragged=None):
    if isinstance(x, tf.SparseTensor) and sparse is not None and not sparse:
        x = sparse_to_dense(x)
    if isinstance(x, tf.RaggedTensor) and ragged is not None and not ragged:
        x = x.to_tensor()
    if dtype is not None:
        dtype = standardize_dtype(dtype)
    if not tf.is_tensor(x):
        if dtype == "bool" or is_int_dtype(dtype):
            # TensorFlow conversion is stricter than other backends, it does not
            # allow ints for bools or floats for ints. We convert without dtype
            # and cast instead.
            x = tf.convert_to_tensor(x)
            return tf.cast(x, dtype)
        return tf.convert_to_tensor(x, dtype=dtype)
    elif dtype is not None and not standardize_dtype(x.dtype) == dtype:
        if isinstance(x, tf.SparseTensor):
            x_shape = x.shape
            x = tf.cast(x, dtype)
            x.set_shape(x_shape)
            return x
        return tf.cast(x, dtype=dtype)
    return x


def convert_to_numpy(x):
    if isinstance(x, tf.SparseTensor):
        x = sparse_to_dense(x)
    elif isinstance(x, tf.IndexedSlices):
        x = tf.convert_to_tensor(x)
    elif isinstance(x, tf.RaggedTensor):
        x = x.to_tensor()
    return np.array(x)


def is_tensor(x):
    return tf.is_tensor(x)


def shape(x):
    """Always return a tuple shape.

    `tf.shape` will return a `tf.Tensor`, which differs from the tuple return
    type on the torch and jax backends. We write our own method instead which
    always returns a tuple, with integer values when the shape is known, and
    tensor values when the shape is unknown (this is tf specific, as dynamic
    shapes do not apply in other backends).
    """
    if isinstance(x, KerasTensor):
        return x.shape
    if not tf.is_tensor(x):
        x = tf.convert_to_tensor(x)
    if x.shape == tf.TensorShape(None):
        raise ValueError(
            "All tensors passed to `ops.shape` must have a statically known "
            f"rank. Received: x={x} with unknown rank."
        )
    shape = x.shape.as_list()
    dynamic = tf.shape(x)
    for i in range(len(shape)):
        if shape[i] is None:
            try:
                shape[i] = dynamic[i]
            except:
                # With RaggedTensors, accessing a ragged dimension will fail,
                # we leave it as None.
                pass
    return tuple(shape)


def cast(x, dtype):
    dtype = standardize_dtype(dtype)
    if isinstance(x, tf.SparseTensor):
        x_shape = x.shape
        x = tf.cast(x, dtype)
        x.set_shape(x_shape)
        return x
    else:
        return tf.cast(x, dtype=dtype)


def compute_output_spec(fn, *args, **kwargs):
    with StatelessScope(), SymbolicScope():
        graph_name = auto_name("scratch_graph")
        with tf.__internal__.FuncGraph(graph_name).as_default():

            def convert_keras_tensor_to_tf(x):
                if isinstance(x, KerasTensor):
                    if x.sparse:
                        return tf.compat.v1.sparse_placeholder(
                            shape=x.shape, dtype=x.dtype
                        )
                    else:
                        return tf.compat.v1.placeholder(
                            shape=x.shape, dtype=x.dtype
                        )
                return x

            args, kwargs = tree.map_structure(
                convert_keras_tensor_to_tf, (args, kwargs)
            )
            tf_out = fn(*args, **kwargs)

            def convert_tf_to_keras_tensor(x):
                if tf.is_tensor(x):
                    return KerasTensor(
                        x.shape, x.dtype, sparse=isinstance(x, tf.SparseTensor)
                    )
                return x

            output_spec = tree.map_structure(convert_tf_to_keras_tensor, tf_out)
    return output_spec


def cond(pred, true_fn, false_fn):
    if isinstance(pred, tf.Variable):
        return tf.cond(pred, true_fn=true_fn, false_fn=false_fn)
    return tf.__internal__.smart_cond.smart_cond(
        pred, true_fn=true_fn, false_fn=false_fn
    )


def vectorized_map(function, elements):
    return tf.vectorized_map(function, elements)


def map(f, xs):
    xs = tree.map_structure(convert_to_tensor, xs)

    def get_fn_output_signature(x):
        out = f(x)
        return tree.map_structure(tf.TensorSpec.from_tensor, out)

    if tree.is_nested(xs):
        input = tree.pack_sequence_as(xs, [x[0] for x in tree.flatten(xs)])
        fn_output_signature = get_fn_output_signature(input)
        return tf.map_fn(f, xs, fn_output_signature=fn_output_signature)
    else:
        fn_output_signature = get_fn_output_signature(xs[0])
        return tf.map_fn(f, xs, fn_output_signature=fn_output_signature)


def scan(f, init, xs=None, length=None, reverse=False, unroll=1):
    # We have reimplemented `scan` to match the behavior of `jax.lax.scan`
    # Ref: tf.scan, jax.lax.scan
    if not callable(f):
        raise TypeError(f"`f` should be a callable. Received: f={f}")
    if not isinstance(unroll, bool):
        if not isinstance(unroll, int) or unroll < 1:
            raise ValueError(
                "`unroll` must be an positive integer or boolean. "
                f"Received: unroll={unroll}"
            )
    if xs is None and length is None:
        raise ValueError("Got no `xs` to scan over and `length` not provided.")

    input_is_sequence = tree.is_nested(xs)
    output_is_sequence = tree.is_nested(init)

    def pack_input(x):
        return tree.pack_sequence_as(xs, x) if input_is_sequence else x[0]

    def pack_output(x):
        return tree.pack_sequence_as(init, x) if output_is_sequence else x[0]

    if xs is None:
        xs_flat = []
        n = int(length)
    else:
        # xs_flat = flatten_input(xs)
        xs_flat = tree.flatten(xs)
        xs_flat = [tf.convert_to_tensor(elem) for elem in xs_flat]
        n = int(length) if length is not None else tf.shape(xs_flat[0])[0]

    # TensorArrays are always flat
    xs_array = [
        tf.TensorArray(
            dtype=x.dtype,
            size=n,
            dynamic_size=False,
            element_shape=x.shape[1:],
            infer_shape=True,
        )
        for x in xs_flat
    ]
    xs_array = [x_a.unstack(x) for x_a, x in zip(xs_array, xs_flat)]

    init_flat = tree.flatten(init)
    carry_flat = [tf.convert_to_tensor(init) for init in init_flat]

    # Store the intermediate values
    # Note: there is a constraint that the output of `f` must have the same
    # shape and dtype as carry (`init`).
    ys_array = [
        tf.TensorArray(
            dtype=carry.dtype,
            size=n,
            dynamic_size=False,
            element_shape=carry.shape,
            infer_shape=True,
        )
        for carry in carry_flat
    ]
    carry_array = [
        tf.TensorArray(
            dtype=carry.dtype,
            size=1,
            dynamic_size=False,
            clear_after_read=False,
            element_shape=carry.shape,
            infer_shape=True,
        )
        for carry in carry_flat
    ]
    carry_array = [
        carry.write(0, c) for (carry, c) in zip(carry_array, carry_flat)
    ]

    def loop_body(i, carry_array, ys_array):
        packed_xs = (
            pack_input([xs.read(i) for xs in xs_array])
            if len(xs_array) > 0
            else None
        )
        packed_carry = pack_output([carry.read(0) for carry in carry_array])

        carry, ys = f(packed_carry, packed_xs)

        if ys is not None:
            flat_ys = tree.flatten(ys)
            ys_array = [ys.write(i, v) for (ys, v) in zip(ys_array, flat_ys)]
        if carry is not None:
            flat_carry = tree.flatten(carry)
            carry_array = [
                carry.write(0, v) for (carry, v) in zip(carry_array, flat_carry)
            ]
        next_i = i + 1 if not reverse else i - 1
        return (next_i, carry_array, ys_array)

    if isinstance(unroll, bool):
        unroll = max(n, 1) if unroll else 1

    _, carry_array, ys_array = tf.while_loop(
        lambda i, _1, _2: i >= 0 if reverse else i < n,
        loop_body,
        (n - 1 if reverse else 0, carry_array, ys_array),
        parallel_iterations=unroll,
    )

    ys_flat = [ys.stack() for ys in ys_array]
    carry_flat = [carry.read(0) for carry in carry_array]
    if xs is not None:
        n_static = xs_flat[0].get_shape().with_rank_at_least(1)[0]
        if not isinstance(n_static, int):
            for x in xs_flat[1:]:
                n_static.assert_is_compatible_with(
                    x.get_shape().with_rank_at_least(1)[0]
                )
        for r in ys_flat:
            r.set_shape(tf.TensorShape(n_static).concatenate(r.get_shape()[1:]))
    return pack_output(carry_flat), pack_output(ys_flat)


def associative_scan(f, elems, reverse=False, axis=0):
    # Implementation is the same as tfp.math.scan_associative
    # with additional checks to ensure similar behavior with jax
    if not callable(f):
        raise TypeError(f"`f` should be a callable. Received: f={f}")
    elems_flat = tree.flatten(elems)
    elems_flat = [tf.convert_to_tensor(elem) for elem in elems_flat]
    if reverse:
        elems_flat = [tf.reverse(elem, [axis]) for elem in elems_flat]

    def _combine(a_flat, b_flat):
        a = tree.pack_sequence_as(elems, a_flat)
        b = tree.pack_sequence_as(elems, b_flat)
        c = f(a, b)
        c_flat = tree.flatten(c)
        return c_flat

    def _get_dim(x):
        return shape(x)[axis]

    # TODO add constant dim check
    num_elems = _get_dim(elems_flat[0])
    if not all(_get_dim(elem) == num_elems for elem in elems_flat[1:]):
        raise ValueError(
            "Array inputs to associative_scan must have the same "
            "first dimension. (saw: {})".format(
                [tf.shape(elem) for elem in elems_flat]
            )
        )

    def _interleave(a, b, axis):
        # [a b c ...] [d e f ...] -> [a d b e c f ...]
        num_elems_a = (
            a.shape[axis] if a.shape[axis] is not None else tf.shape(a)[axis]
        )
        num_elems_b = (
            b.shape[axis] if b.shape[axis] is not None else tf.shape(b)[axis]
        )

        # Note that interleaving implies rank(a)==rank(b).
        axis = tf.where(axis >= 0, axis, tf.rank(a) + axis)
        axis = (
            int(axis)  # Avoid ndarray values.
            if tf.get_static_value(axis) is not None
            else axis
        )

        def _interleave_with_b(a):
            a_shape = a.shape.as_list() if a.shape.rank is not None else None
            if isinstance(axis, int):
                b_shape_at_axis = b.shape[axis]
                num_elems_b_static = tf.get_static_value(
                    b_shape_at_axis
                    if b_shape_at_axis is not None
                    else tf.shape(b)[axis]
                )
            else:
                num_elems_b_static = None

            if (
                isinstance(axis, int)
                and a_shape is not None
                and num_elems_b_static is not None
                and all(
                    d is not None for i, d in enumerate(a_shape) if i != axis
                )
            ):
                new_shape = (
                    a_shape[:axis]
                    + [2 * num_elems_b_static]
                    + a_shape[axis + 1 :]
                )
            else:
                new_shape = tf.concat(
                    [
                        tf.shape(a)[:axis],
                        [2 * num_elems_b],
                        tf.shape(a)[axis + 1 :],
                    ],
                    axis=0,
                )
            return tf.reshape(
                # Work around lack of support for Tensor axes in
                # `tf.stack` by using `concat` and `expand_dims` instead.
                tf.concat(
                    [
                        tf.expand_dims(a, axis=axis + 1),
                        tf.expand_dims(b, axis=axis + 1),
                    ],
                    axis=axis + 1,
                ),
                new_shape,
            )

        return tf.cond(
            tf.equal(num_elems_a, num_elems_b + 1),
            lambda: tf.concat(
                [
                    _interleave_with_b(
                        tf.gather(
                            a,
                            tf.range(0, num_elems_a - 1),
                            axis=axis,
                        )
                    ),
                    tf.gather(
                        a,
                        tf.range(num_elems_a - 1, num_elems_a),
                        axis=axis,
                    ),
                ],
                axis=axis,
            ),
            lambda: _interleave_with_b(a),
        )

    # Dynamic-shape helper: resolve negative slice indices against the
    # actual (runtime) axis length using tf.shape, so that slicing works
    # inside tf.cond branches where the axis dim is statically unknown.
    def _dynamic_slice(x, start=0, stop=None, step=1):
        axis_len = tf.shape(x)[axis]
        if start is None:
            start = 0
        elif start < 0:
            start = start + axis_len
        if stop is None:
            stop = axis_len
        elif stop < 0:
            stop = stop + axis_len
        return tf.gather(x, tf.range(start, stop, delta=step), axis=axis)

    def _process_recursive_result(
        odd_elems, elems_level, is_even_length, use_dynamic_slicing=False
    ):
        """
        Helper function to process the result of a recursive _scan call.

        Takes the recursively scanned odd-indexed elements and applies the
        post-processing logic (combining, concatenation, and interleaving)
        that is common to both static and symbolic execution paths.

        Args:
            odd_elems: Result from _scan(reduced_elems)
            elems_level: Elements at the current level
            is_even_length: Boolean or symbolic tensor
                indicating if elem_length % 2 == 0
            use_dynamic_slicing: If True, uses _dynamic_slice for
                negative indices instead of slice_along_axis.
                Required when shapes are unknown at trace time.

        Returns:
            The final scanned result after interleaving even and odd
            elements.
        """
        _sl = (
            _dynamic_slice
            if use_dynamic_slicing
            else (
                lambda x, start=0, stop=None, step=1: slice_along_axis(
                    x, start, stop, step, axis=axis
                )
            )
        )

        def _get_even_results():
            return _combine(
                [_sl(odd_elem, 0, -1) for odd_elem in odd_elems],
                [_sl(elem, 2, None, 2) for elem in elems_level],
            )

        def _get_odd_results():
            return _combine(
                [odd_elem for odd_elem in odd_elems],
                [_sl(elem, 2, None, 2) for elem in elems_level],
            )

        # For Python-level conditionals (static), is_even_length is a bool
        # For symbolic conditionals, is_even_length is a Tensor,
        # handled via tf.cond
        if isinstance(is_even_length, bool):
            results = (
                _get_even_results() if is_even_length else _get_odd_results()
            )
        else:
            results = tf.cond(
                is_even_length, _get_even_results, _get_odd_results
            )

        even_elems = [
            tf.concat([_sl(elem, 0, 1), result], axis=axis)
            for (elem, result) in zip(elems_level, results)
        ]
        return list(
            builtins.map(
                lambda a, b: _interleave(a, b, axis=axis),
                even_elems,
                odd_elems,
            )
        )

    # Max recursion depth for dynamic-shape fallback. Supports sequences upto
    # 2^48 elements.
    _max_num_levels = 48

    def _scan(elems):
        elem_length = _get_dim(elems[0])

        # Try to get the static value of elem_length to avoid infinite
        # recursion in tf.function when elem_length is symbolic.
        static_elem_length = tf.get_static_value(elem_length)

        # Early return before doing any slicing work.
        if static_elem_length is not None and int(static_elem_length) < 2:
            return elems

        a = [slice_along_axis(elem, 0, -1, step=2, axis=axis) for elem in elems]
        b = [
            slice_along_axis(elem, 1, None, step=2, axis=axis) for elem in elems
        ]
        reduced_elems = _combine(a, b)

        def _handle_base_case_elem_length_two():
            return [
                tf.concat(
                    [slice_along_axis(elem, 0, 1, axis=axis), reduced_elem],
                    axis=axis,
                )
                for (reduced_elem, elem) in zip(reduced_elems, elems)
            ]

        def _handle_base_case_elem_length_three():
            reduced_reduced_elems = _combine(
                reduced_elems,
                [slice_along_axis(elem, 2, 3, axis=axis) for elem in elems],
            )
            return [
                tf.concat(
                    [
                        slice_along_axis(elem, 0, 1, axis=axis),
                        reduced_elem,
                        reduced_reduced_elem,
                    ],
                    axis=axis,
                )
                for (reduced_reduced_elem, reduced_elem, elem) in zip(
                    reduced_reduced_elems, reduced_elems, elems
                )
            ]

        # If we have a static value, use Python conditionals to avoid
        # tracing the recursive case unnecessarily
        if static_elem_length is not None:
            static_elem_length = int(static_elem_length)
            if static_elem_length == 2:
                return _handle_base_case_elem_length_two()
            elif static_elem_length == 3:
                return _handle_base_case_elem_length_three()
            else:
                # Recursive case with static length
                odd_elems = _scan(reduced_elems)
                # Use Python bool for static length
                is_even = static_elem_length % 2 == 0
                return _process_recursive_result(odd_elems, elems, is_even)

        # Fallback for symbolic lengths: use TF-cond recursion (following
        # the TFP scan_associative pattern) with a level counter to bound
        # the Python-level recursion depth. At each level tf.cond chooses
        # between the base case and the recursive case at runtime.
        return tf.cond(
            tf.less(elem_length, 2),
            lambda: elems,
            lambda: _scan_dynamic(_max_num_levels, elems),
        )

    def _scan_dynamic(level, elems):
        """Recursive associative scan for symbolic (dynamic) lengths.

        Uses a level counter to bound the Python-level unrolling depth,
        mirroring the approach from tfp.math.scan_associative.  Each
        level is guarded by tf.cond so only the needed depth executes.
        """
        elem_length = tf.shape(elems[0])[axis]

        a = [_dynamic_slice(elem, 0, -1, step=2) for elem in elems]
        b = [_dynamic_slice(elem, 1, None, step=2) for elem in elems]
        reduced_elems = _combine(a, b)

        def _handle_base_case_two():
            return [
                tf.concat([_dynamic_slice(elem, 0, 1), reduced_elem], axis=axis)
                for (reduced_elem, elem) in zip(reduced_elems, elems)
            ]

        def _handle_base_case_three():
            reduced_reduced_elems = _combine(
                reduced_elems,
                [_dynamic_slice(elem, 2, 3) for elem in elems],
            )
            return [
                tf.concat(
                    [
                        _dynamic_slice(elem, 0, 1),
                        reduced_elem,
                        reduced_reduced_elem,
                    ],
                    axis=axis,
                )
                for (reduced_reduced_elem, reduced_elem, elem) in zip(
                    reduced_reduced_elems, reduced_elems, elems
                )
            ]

        at_base_case = tf.logical_or(
            tf.equal(elem_length, 2), tf.equal(elem_length, 3)
        )
        base_value = lambda: tf.cond(
            tf.equal(elem_length, 2),
            _handle_base_case_two,
            _handle_base_case_three,
        )

        if level <= 0:
            return base_value()

        def _recursive_case():
            odd_elems = _scan_dynamic(level - 1, reduced_elems)
            is_even = tf.equal(elem_length % 2, 0)
            return _process_recursive_result(
                odd_elems, elems, is_even, use_dynamic_slicing=True
            )

        return tf.cond(at_base_case, base_value, _recursive_case)

    scans = _scan(elems_flat)
    if reverse:
        scans = [tf.reverse(scanned, [axis]) for scanned in scans]

    return tree.pack_sequence_as(elems, scans)


def scatter(indices, values, shape):
    return tf.scatter_nd(indices, values, shape)


def scatter_update(inputs, indices, updates, reduction=None):
    if reduction is None:
        return tf.tensor_scatter_nd_update(inputs, indices, updates)
    elif reduction == "add":
        return tf.tensor_scatter_nd_add(inputs, indices, updates)
    elif reduction == "max":
        return tf.tensor_scatter_nd_max(inputs, indices, updates)
    elif reduction == "min":
        return tf.tensor_scatter_nd_min(inputs, indices, updates)
    elif reduction == "mul":
        # TensorFlow doesn't have tensor_scatter_nd_mul, implement manually
        # Use while_loop to handle both scalar and slice updates correctly
        num_updates = tf.shape(indices)[0]

        def body(i, result):
            idx = indices[i : i + 1]  # Shape (1, index_depth)
            current = tf.gather_nd(result, idx)  # Shape (1, *slice_shape)
            new_value = (
                current * updates[i]
            )  # Maintains shape (1, *slice_shape)
            return i + 1, tf.tensor_scatter_nd_update(result, idx, new_value)

        _, result = tf.while_loop(
            lambda i, _: i < num_updates,
            body,
            [0, inputs],
        )
        return result
    else:
        raise ValueError(f"Unsupported reduction: {reduction}")


def slice(inputs, start_indices, shape):
    return tf.slice(inputs, start_indices, shape)


def slice_update(inputs, start_indices, updates):
    return dynamic_update_slice(inputs, updates, start_indices)


def switch(index, branches, *operands):
    index = convert_to_tensor(index, "int32")
    index = tf.clip_by_value(index, 0, len(branches) - 1)

    # Workaround to deal with python closures. More details:
    # https://github.com/tensorflow/tensorflow/issues/8776#issuecomment-311383887
    def gen_fn(i):
        return lambda: branches[i](*operands)

    branch_fns = [gen_fn(i) for i in range(len(branches))]
    return tf.switch_case(index, branch_fns)


def while_loop(
    cond,
    body,
    loop_vars,
    maximum_iterations=None,
):
    is_tuple = isinstance(loop_vars, (tuple, list))
    loop_vars = tuple(loop_vars) if is_tuple else (loop_vars,)

    def _body(*args):
        outputs = body(*args)
        return tuple(outputs) if is_tuple else (outputs,)

    outputs = tf.while_loop(
        cond,
        _body,
        loop_vars,
        maximum_iterations=maximum_iterations,
    )
    return outputs if is_tuple else outputs[0]


def fori_loop(lower, upper, body_fun, init_val):
    return tf.while_loop(
        lambda i, val: i < upper,
        lambda i, val: (i + 1, body_fun(i, val)),
        (lower, init_val),
    )[1]


def stop_gradient(variable):
    return tf.stop_gradient(variable)


def unstack(x, num=None, axis=0):
    return tf.unstack(x, num=num, axis=axis)


def random_seed_dtype():
    # tensorflow random operation only works on int32/int64, not uint32.
    return "int64"


def custom_gradient(fun):
    return tf.custom_gradient(f=fun)


def remat(f):
    """Implementation of rematerialization.

    Args:
        f: The function or operation to rematerialize.
    Returns:
        A function wrapping f that defines a custom gradient, which
        recomputes f on the backwards pass of a gradient call.
    """
    return tf.recompute_grad(f)


class name_scope(base_name_scope):
    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)
        self._tf_name_scope = tf.name_scope(name)

    def __enter__(self):
        name_scope_stack = global_state.get_global_attribute(
            "name_scope_stack", default=[], set_to_default=True
        )
        if self.deduplicate and name_scope_stack:
            parent_caller = name_scope_stack[-1].caller
            parent_name = name_scope_stack[-1].name
            if (
                self.caller is not None
                and self.caller is parent_caller
                and self.name == parent_name
            ):
                return self
        name_scope_stack.append(self)
        self._pop_on_exit = True
        self._tf_name_scope.__enter__()
        return self

    def __exit__(self, *args, **kwargs):
        super().__exit__(*args, **kwargs)
        if self._pop_on_exit:
            self._tf_name_scope.__exit__(*args, **kwargs)


def device_scope(device_name):
    return tf.device(device_name)



import itertools

import numpy as np
import openvino.opset16 as ov_opset
from openvino import Type

from keras.src import backend
from keras.src.backend.openvino.core import DTYPES_MAX
from keras.src.backend.openvino.core import DTYPES_MIN
from keras.src.backend.openvino.core import OpenVINOKerasTensor
from keras.src.backend.openvino.core import cast
from keras.src.backend.openvino.core import convert_to_numpy
from keras.src.backend.openvino.core import convert_to_tensor
from keras.src.backend.openvino.core import get_ov_output
from keras.src.backend.openvino.random import _random_normal
from keras.src.random.seed_generator import draw_seed

AFFINE_TRANSFORM_INTERPOLATIONS = {"nearest": 0, "bilinear": 1}
AFFINE_TRANSFORM_FILL_MODES = {
    "constant",
    "nearest",
    "wrap",
    "mirror",
    "reflect",
}

SCALE_AND_TRANSLATE_METHODS = {
    "linear",
    "bilinear",
    "trilinear",
    "cubic",
    "bicubic",
    "tricubic",
    "lanczos3",
    "lanczos5",
}
RESIZE_INTERPOLATIONS = {
    "bilinear": "linear",
    "nearest": "nearest",
    "bicubic": "cubic",
}
# Lanczos methods are handled via _ov_scale_and_translate
# instead of ov_opset.interpolate, which does not support them natively.
LANCZOS_INTERPOLATIONS = {
    "lanczos3": "lanczos3",
    "lanczos5": "lanczos5",
}
ALL_RESIZE_INTERPOLATIONS = tuple(RESIZE_INTERPOLATIONS) + tuple(
    LANCZOS_INTERPOLATIONS
)


def rgb_to_grayscale(images, data_format=None):
    images = get_ov_output(images)
    data_format = backend.standardize_data_format(data_format)
    if images.get_partial_shape().rank not in (3, 4):
        raise ValueError(
            "Invalid images rank: expected rank 3 (single image) "
            "or rank 4 (batch of images). Received input with shape: "
            f"images.shape={images.shape}"
        )
    channel_axis = -3 if data_format == "channels_first" else -1
    if images.shape[channel_axis] not in (1, 3):
        raise ValueError(
            "Invalid channel size: expected 3 (RGB) or 1 (Grayscale). "
            f"Received input with shape: images.shape={images.shape}"
        )

    if images.shape[channel_axis] == 3:
        original_type = images.get_element_type()
        rgb_weights = ov_opset.constant(
            [0.2989, 0.5870, 0.1140], dtype=original_type
        ).output(0)
        if data_format == "channels_first":
            rgb_weights = ov_opset.unsqueeze(rgb_weights, axes=[-2, -1]).output(
                0
            )
        grayscales = ov_opset.multiply(images, rgb_weights).output(0)
        grayscales = ov_opset.reduce_sum(
            grayscales, reduction_axes=[channel_axis]
        ).output(0)
        grayscales = ov_opset.unsqueeze(grayscales, axes=[channel_axis]).output(
            0
        )
        if grayscales.get_element_type() != original_type:
            # Type of grayscales may be changed after unsqueeze, so we need to
            # convert it back to the original type.
            grayscales = ov_opset.convert(grayscales, original_type).output(0)

    return OpenVINOKerasTensor(grayscales)


def rgb_to_hsv(images, data_format=None):
    dtype = images.dtype
    images = get_ov_output(images)
    ov_type = images.get_element_type()
    data_format = backend.standardize_data_format(data_format)
    channels_axis = -1 if data_format == "channels_last" else -3
    if len(images.shape) not in (3, 4):
        raise ValueError(
            "Invalid images rank: expected rank 3 (single image) "
            "or rank 4 (batch of images). Received input with shape: "
            f"images.shape={images.shape}"
        )
    if not backend.is_float_dtype(dtype):
        raise ValueError(
            "Invalid images dtype: expected float dtype. "
            f"Received: images.dtype={dtype}"
        )
    eps = ov_opset.constant(backend.epsilon(), dtype=ov_type).output(0)
    images = ov_opset.select(
        ov_opset.less(ov_opset.abs(images), eps),
        ov_opset.constant(0.0, dtype=ov_type),
        images,
    ).output(0)
    rgb_channels = ov_opset.split(images, axis=channels_axis, num_splits=3)
    r, g, b = (
        rgb_channels.output(0),
        rgb_channels.output(1),
        rgb_channels.output(2),
    )

    def rgb_planes_to_hsv_planes(r, g, b):
        value = ov_opset.maximum(ov_opset.maximum(r, g), b).output(0)
        minimum = ov_opset.minimum(ov_opset.minimum(r, g), b).output(0)
        range_ = ov_opset.subtract(value, minimum).output(0)

        safe_value = ov_opset.select(
            ov_opset.greater(value, ov_opset.constant(0.0, dtype=ov_type)),
            value,
            ov_opset.constant(1.0, dtype=ov_type),
        ).output(0)
        safe_range = ov_opset.select(
            ov_opset.greater(range_, ov_opset.constant(0.0, dtype=ov_type)),
            range_,
            ov_opset.constant(1.0, dtype=ov_type),
        ).output(0)

        saturation = ov_opset.select(
            ov_opset.greater(value, ov_opset.constant(0.0, dtype=ov_type)),
            ov_opset.divide(range_, safe_value),
            ov_opset.constant(0.0, dtype=ov_type),
        ).output(0)
        norm = ov_opset.divide(
            ov_opset.constant(1.0, dtype=ov_type),
            ov_opset.multiply(
                ov_opset.constant(6.0, dtype=ov_type), safe_range
            ),
        ).output(0)

        hue = ov_opset.select(
            ov_opset.equal(value, g),
            ov_opset.add(
                ov_opset.multiply(norm, ov_opset.subtract(b, r)),
                ov_opset.constant(2.0 / 6.0, dtype=ov_type),
            ),
            ov_opset.add(
                ov_opset.multiply(norm, ov_opset.subtract(r, g)),
                ov_opset.constant(4.0 / 6.0, dtype=ov_type),
            ),
        ).output(0)
        hue = ov_opset.select(
            ov_opset.equal(value, r),
            ov_opset.multiply(norm, ov_opset.subtract(g, b)),
            hue,
        ).output(0)
        hue = ov_opset.select(
            ov_opset.greater(range_, ov_opset.constant(0.0, dtype=ov_type)),
            hue,
            ov_opset.constant(0.0, dtype=ov_type),
        ).output(0)
        hue = ov_opset.add(
            hue,
            ov_opset.convert(
                ov_opset.less(hue, ov_opset.constant(0.0, dtype=ov_type)),
                ov_type,
            ),
        ).output(0)
        return hue, saturation, value

    images = ov_opset.concat(
        rgb_planes_to_hsv_planes(r, g, b), axis=channels_axis
    ).output(0)
    return OpenVINOKerasTensor(images)


def hsv_to_rgb(images, data_format=None):
    dtype = images.dtype
    images = get_ov_output(images)
    ov_type = images.get_element_type()
    data_format = backend.standardize_data_format(data_format)
    channels_axis = -1 if data_format == "channels_last" else -3
    if len(images.shape) not in (3, 4):
        raise ValueError(
            "Invalid images rank: expected rank 3 (single image) "
            "or rank 4 (batch of images). Received input with shape: "
            f"images.shape={images.shape}"
        )
    if not backend.is_float_dtype(dtype):
        raise ValueError(
            "Invalid images dtype: expected float dtype. "
            f"Received: images.dtype={dtype}"
        )
    hsv_channels = ov_opset.split(images, axis=channels_axis, num_splits=3)
    hue, saturation, value = (
        hsv_channels.output(0),
        hsv_channels.output(1),
        hsv_channels.output(2),
    )

    def hsv_planes_to_rgb_planes(hue, saturation, value):
        def channel_value(channel_delta, one_minus_saturation):
            return ov_opset.multiply(
                value,
                ov_opset.add(
                    one_minus_saturation,
                    ov_opset.multiply(saturation, channel_delta),
                ),
            )

        dh = ov_opset.multiply(
            ov_opset.mod(hue, ov_opset.constant(1.0, dtype=ov_type)),
            ov_opset.constant(6.0, dtype=ov_type),
        ).output(0)
        one_const = ov_opset.constant(1.0, dtype=ov_type).output(0)
        two_const = ov_opset.constant(2.0, dtype=ov_type).output(0)
        three_const = ov_opset.constant(3.0, dtype=ov_type).output(0)
        four_const = ov_opset.constant(4.0, dtype=ov_type).output(0)
        dr = ov_opset.subtract(
            ov_opset.abs(ov_opset.subtract(dh, three_const)), one_const
        ).output(0)
        dr = ov_opset.clamp(dr, 0.0, 1.0).output(0)
        dg = ov_opset.subtract(
            two_const, ov_opset.abs(ov_opset.subtract(dh, two_const))
        ).output(0)
        dg = ov_opset.clamp(dg, 0.0, 1.0).output(0)
        db = ov_opset.subtract(
            two_const, ov_opset.abs(ov_opset.subtract(dh, four_const))
        ).output(0)
        db = ov_opset.clamp(db, 0.0, 1.0).output(0)
        one_minus_saturation = ov_opset.subtract(one_const, saturation).output(
            0
        )

        red = channel_value(dr, one_minus_saturation)
        green = channel_value(dg, one_minus_saturation)
        blue = channel_value(db, one_minus_saturation)
        return red, green, blue

    images = ov_opset.concat(
        hsv_planes_to_rgb_planes(hue, saturation, value), axis=channels_axis
    ).output(0)
    return OpenVINOKerasTensor(images)


def resize(
    images,
    size,
    interpolation="bilinear",
    antialias=False,
    crop_to_aspect_ratio=False,
    pad_to_aspect_ratio=False,
    fill_mode="constant",
    fill_value=0.0,
    data_format="channels_last",
):
    data_format = backend.standardize_data_format(data_format)
    if interpolation not in ALL_RESIZE_INTERPOLATIONS:
        raise ValueError(
            "Invalid value for argument `interpolation`. Expected of one "
            f"{ALL_RESIZE_INTERPOLATIONS}. Received: "
            f"interpolation={interpolation}"
        )
    if fill_mode != "constant":
        raise ValueError(
            "Invalid value for argument `fill_mode`. Only `'constant'` "
            f"is supported. Received: fill_mode={fill_mode}"
        )
    if pad_to_aspect_ratio and crop_to_aspect_ratio:
        raise ValueError(
            "Only one of `pad_to_aspect_ratio` & `crop_to_aspect_ratio` "
            "can be `True`."
        )
    if not len(size) == 2:
        raise ValueError(
            "Argument `size` must be a tuple of two elements "
            f"(height, width). Received: size={size}"
        )

    target_height, target_width = tuple(size)
    images = get_ov_output(images)
    rank = len(images.get_partial_shape())
    if rank not in (3, 4):
        raise ValueError(
            "Invalid images rank: expected rank 3 (single image) "
            "or rank 4 (batch of images). Received input with shape: "
            f"images.shape={images.get_partial_shape()}"
        )

    if data_format == "channels_last":
        height_axis, width_axis = (-3, -2)
    else:
        height_axis, width_axis = (-2, -1)

    def _gather_dim(shape_tensor, axis):
        axis_node = ov_opset.constant([axis], Type.i32).output(0)
        axis0 = ov_opset.constant(0, Type.i32).output(0)
        return ov_opset.gather(shape_tensor, axis_node, axis0).output(0)

    def _floor_div_int(numerator, denominator):
        numerator_f = ov_opset.convert(numerator, Type.f32).output(0)
        denominator_f = ov_opset.convert(denominator, Type.f32).output(0)
        quotient = ov_opset.divide(numerator_f, denominator_f).output(0)
        floored = ov_opset.floor(quotient).output(0)
        return ov_opset.convert(floored, Type.i32).output(0)

    def _concat_scalars(nodes):
        return ov_opset.concat(nodes, axis=0).output(0)

    shape_node = ov_opset.shape_of(images, Type.i32).output(0)
    height_axis_index = height_axis % rank
    width_axis_index = width_axis % rank
    height = _gather_dim(shape_node, height_axis_index)
    width = _gather_dim(shape_node, width_axis_index)

    target_height_node = ov_opset.constant([target_height], Type.i32).output(0)
    target_width_node = ov_opset.constant([target_width], Type.i32).output(0)
    one_i32 = ov_opset.constant([1], Type.i32).output(0)
    zero_i32 = ov_opset.constant([0], Type.i32).output(0)

    if crop_to_aspect_ratio:
        crop_height = _floor_div_int(
            ov_opset.multiply(width, target_height_node).output(0),
            target_width_node,
        )
        crop_height = ov_opset.minimum(height, crop_height).output(0)
        crop_height = ov_opset.maximum(one_i32, crop_height).output(0)

        crop_width = _floor_div_int(
            ov_opset.multiply(height, target_width_node).output(0),
            target_height_node,
        )
        crop_width = ov_opset.minimum(width, crop_width).output(0)
        crop_width = ov_opset.maximum(one_i32, crop_width).output(0)

        crop_box_hstart = _floor_div_int(
            ov_opset.subtract(height, crop_height).output(0),
            ov_opset.constant([2], Type.i32).output(0),
        )
        crop_box_wstart = _floor_div_int(
            ov_opset.subtract(width, crop_width).output(0),
            ov_opset.constant([2], Type.i32).output(0),
        )

        crop_box_hend = ov_opset.add(crop_box_hstart, crop_height).output(0)
        crop_box_wend = ov_opset.add(crop_box_wstart, crop_width).output(0)

        begin_parts = []
        end_parts = []
        begin_mask = [1] * rank
        end_mask = [1] * rank
        for axis in range(rank):
            if axis == height_axis_index:
                begin_parts.append(crop_box_hstart)
                end_parts.append(crop_box_hend)
                begin_mask[axis] = 0
                end_mask[axis] = 0
            elif axis == width_axis_index:
                begin_parts.append(crop_box_wstart)
                end_parts.append(crop_box_wend)
                begin_mask[axis] = 0
                end_mask[axis] = 0
            else:
                begin_parts.append(zero_i32)
                end_parts.append(zero_i32)

        images = ov_opset.strided_slice(
            data=images,
            begin=_concat_scalars(begin_parts),
            end=_concat_scalars(end_parts),
            strides=ov_opset.constant([1] * rank, Type.i32).output(0),
            begin_mask=begin_mask,
            end_mask=end_mask,
        ).output(0)
    elif pad_to_aspect_ratio:
        pad_height = _floor_div_int(
            ov_opset.multiply(width, target_height_node).output(0),
            target_width_node,
        )
        pad_height = ov_opset.maximum(height, pad_height).output(0)

        pad_width = _floor_div_int(
            ov_opset.multiply(height, target_width_node).output(0),
            target_height_node,
        )
        pad_width = ov_opset.maximum(width, pad_width).output(0)

        img_box_hstart = _floor_div_int(
            ov_opset.subtract(pad_height, height).output(0),
            ov_opset.constant([2], Type.i32).output(0),
        )
        img_box_wstart = _floor_div_int(
            ov_opset.subtract(pad_width, width).output(0),
            ov_opset.constant([2], Type.i32).output(0),
        )

        pads_begin_parts = []
        pads_end_parts = []
        for axis in range(rank):
            if axis == height_axis_index:
                pads_begin_parts.append(img_box_hstart)
                pads_end_parts.append(img_box_hstart)
            elif axis == width_axis_index:
                pads_begin_parts.append(img_box_wstart)
                pads_end_parts.append(img_box_wstart)
            else:
                pads_begin_parts.append(zero_i32)
                pads_end_parts.append(zero_i32)

        fill_value = ov_opset.constant(
            fill_value, images.get_element_type()
        ).output(0)
        images = ov_opset.pad(
            images,
            _concat_scalars(pads_begin_parts),
            _concat_scalars(pads_end_parts),
            "constant",
            fill_value,
        ).output(0)

    if interpolation in LANCZOS_INTERPOLATIONS:
        # routed through the weight-matrix resampler used by
        # scale_and_translate, which has lanczos kernel support.
        post_crop_shape = ov_opset.shape_of(images, Type.i32).output(0)
        src_h = ov_opset.convert(
            _gather_dim(post_crop_shape, height_axis_index), Type.f32
        ).output(0)
        src_w = ov_opset.convert(
            _gather_dim(post_crop_shape, width_axis_index), Type.f32
        ).output(0)
        scale_h = ov_opset.divide(
            ov_opset.constant(float(target_height), Type.f32).output(0),
            src_h,
        ).output(0)
        scale_w = ov_opset.divide(
            ov_opset.constant(float(target_width), Type.f32).output(0),
            src_w,
        ).output(0)
        scale_ov = ov_opset.concat([scale_h, scale_w], axis=0).output(0)
        translation_ov = ov_opset.constant([0.0, 0.0], dtype=Type.f32).output(0)
        output_shape_map = {
            height_axis_index: target_height,
            width_axis_index: target_width,
        }
        spatial_dims = [height_axis_index, width_axis_index]
        kernel = _ov_kernels[LANCZOS_INTERPOLATIONS[interpolation]]
        return _ov_scale_and_translate(
            images,
            output_shape_map,
            spatial_dims,
            scale_ov,
            translation_ov,
            kernel,
            antialias,
        )

    axes = [height_axis % rank, width_axis % rank]
    size = ov_opset.constant([target_height, target_width], Type.i32).output(0)
    axes = ov_opset.constant(axes, Type.i32).output(0)

    original_type = images.get_element_type()
    supported_types = {
        Type.f32,
        Type.f16,
        Type.bf16,
        Type.i8,
        Type.u8,
        Type.i32,
        Type.i64,
    }
    should_round_before_cast = False
    should_adjust_uint8_bicubic = False
    if original_type == Type.u8 and interpolation != "nearest":
        images = ov_opset.convert(images, Type.f32).output(0)
        should_round_before_cast = True
        if interpolation == "bicubic":
            should_adjust_uint8_bicubic = True
    elif original_type not in supported_types:
        images = ov_opset.convert(images, Type.f32).output(0)

    interpolate_kwargs = {
        "mode": RESIZE_INTERPOLATIONS[interpolation],
        "shape_calculation_mode": "sizes",
        "antialias": antialias,
        "axes": axes,
    }
    if interpolation == "nearest":
        interpolate_kwargs["coordinate_transformation_mode"] = (
            "tf_half_pixel_for_nn"
        )
        interpolate_kwargs["nearest_mode"] = "simple"
    elif interpolation == "bicubic":
        interpolate_kwargs["coordinate_transformation_mode"] = "half_pixel"
        interpolate_kwargs["cube_coeff"] = -0.5
    else:
        interpolate_kwargs["coordinate_transformation_mode"] = "half_pixel"

    resized = ov_opset.interpolate(images, size, **interpolate_kwargs).output(0)

    if should_round_before_cast:
        resized = ov_opset.round(resized, "half_to_even").output(0)
        if should_adjust_uint8_bicubic:
            # Match TensorFlow/OpenCV-style uint8 bicubic behavior more closely
            # by nudging non-extreme values toward mid-range before clamping.
            low_mask = ov_opset.logical_and(
                ov_opset.greater(resized, ov_opset.constant(0.0, Type.f32)),
                ov_opset.less(resized, ov_opset.constant(128.0, Type.f32)),
            ).output(0)
            high_mask = ov_opset.logical_and(
                ov_opset.greater_equal(
                    resized, ov_opset.constant(128.0, Type.f32)
                ),
                ov_opset.less(resized, ov_opset.constant(255.0, Type.f32)),
            ).output(0)
            plus_one = ov_opset.add(
                resized,
                ov_opset.convert(low_mask, Type.f32),
            ).output(0)
            resized = ov_opset.subtract(
                plus_one,
                ov_opset.convert(high_mask, Type.f32),
            ).output(0)
        resized = ov_opset.clamp(resized, 0.0, 255.0).output(0)
    if resized.get_element_type() != original_type:
        resized = ov_opset.convert(resized, original_type).output(0)
    return OpenVINOKerasTensor(resized)


def _ov_build_affine_coords(images_ov, transform_ov):
    """Build transformed coordinates for affine_transform using OV opsets.

    images_ov: [B, H, W, C] f32
    transform_ov: [B, 8] f32 (TF-convention flat transform)
    Returns coords: [4, B, H, W, C] f32 — (batch, row, col, chan) per pixel.
    """
    shape_node = ov_opset.shape_of(images_ov, output_type=Type.i32).output(0)
    axis0 = ov_opset.constant(0, Type.i32).output(0)

    def dim(i):
        return ov_opset.gather(
            shape_node, ov_opset.constant(i, Type.i32).output(0), axis0
        ).output(0)

    B = dim(0)
    H = dim(1)
    W = dim(2)
    C = dim(3)

    H_f = ov_opset.convert(H, Type.f32).output(0)
    W_f = ov_opset.convert(W, Type.f32).output(0)
    C_f = ov_opset.convert(C, Type.f32).output(0)
    zero_f = ov_opset.constant(0.0, Type.f32).output(0)
    one_f = ov_opset.constant(1.0, Type.f32).output(0)
    r_h = ov_opset.range(zero_f, H_f, one_f, output_type=Type.f32).output(0)
    r_w = ov_opset.range(zero_f, W_f, one_f, output_type=Type.f32).output(0)
    r_c = ov_opset.range(zero_f, C_f, one_f, output_type=Type.f32).output(0)

    def s1d(scalar):
        return ov_opset.reshape(
            scalar, ov_opset.constant([1], Type.i32).output(0), False
        ).output(0)

    hwc_shape = ov_opset.concat([s1d(H), s1d(W), s1d(C)], axis=0).output(0)
    grid_h = ov_opset.broadcast(
        ov_opset.reshape(
            r_h,
            ov_opset.concat(
                [s1d(H), ov_opset.constant([1, 1], Type.i32).output(0)], axis=0
            ).output(0),
            False,
        ).output(0),
        hwc_shape,
    ).output(0)
    grid_w = ov_opset.broadcast(
        ov_opset.reshape(
            r_w,
            ov_opset.concat(
                [
                    ov_opset.constant([1], Type.i32).output(0),
                    s1d(W),
                    ov_opset.constant([1], Type.i32).output(0),
                ],
                axis=0,
            ).output(0),
            False,
        ).output(0),
        hwc_shape,
    ).output(0)
    grid_c = ov_opset.broadcast(
        ov_opset.reshape(
            r_c,
            ov_opset.concat(
                [ov_opset.constant([1, 1], Type.i32).output(0), s1d(C)], axis=0
            ).output(0),
            False,
        ).output(0),
        hwc_shape,
    ).output(0)

    neg1 = ov_opset.constant([-1], Type.i32).output(0)
    flat_h = ov_opset.reshape(grid_h, neg1, False).output(0)
    flat_w = ov_opset.reshape(grid_w, neg1, False).output(0)
    one_vec = ov_opset.broadcast(
        ov_opset.constant(1.0, Type.f32).output(0),
        ov_opset.shape_of(flat_h, output_type=Type.i32).output(0),
    ).output(0)
    pts = ov_opset.concat(
        [
            ov_opset.unsqueeze(flat_h, axes=[1]).output(0),
            ov_opset.unsqueeze(flat_w, axes=[1]).output(0),
            ov_opset.unsqueeze(one_vec, axes=[1]).output(0),
        ],
        axis=1,
    ).output(0)

    def t_col(i):
        return ov_opset.gather(
            transform_ov,
            ov_opset.constant(i, Type.i32).output(0),
            ov_opset.constant(1, Type.i32).output(0),
        ).output(0)  # [B]

    a0, a1, a2 = t_col(0), t_col(1), t_col(2)
    b0, b1, b2 = t_col(3), t_col(4), t_col(5)
    ones_b = ov_opset.broadcast(
        ov_opset.constant(1.0, Type.f32).output(0),
        ov_opset.shape_of(a0, output_type=Type.i32).output(0),
    ).output(0)
    zeros_b = ov_opset.broadcast(
        ov_opset.constant(0.0, Type.f32).output(0),
        ov_opset.shape_of(a0, output_type=Type.i32).output(0),
    ).output(0)

    # Numpy reference matrix (after swap + zero-out offset col):
    #   T = [[b1, a1, 0], [b0, a0, 0], [0, 0, 1]]   shape [3,3]
    # offset = [b2, a2, 0]
    # coords = pts @ T + offset  (right-multiply)
    def make_col(c0, c1, c2):
        col = ov_opset.concat(
            [
                ov_opset.unsqueeze(c0, axes=[1]).output(0),
                ov_opset.unsqueeze(c1, axes=[1]).output(0),
                ov_opset.unsqueeze(c2, axes=[1]).output(0),
            ],
            axis=1,
        ).output(0)  # [B, 3]
        return ov_opset.unsqueeze(col, axes=[2]).output(0)  # [B, 3, 1]

    # T as columns: col0=[b1,b0,0], col1=[a1,a0,0], col2=[0,0,1]
    col0 = make_col(b1, b0, zeros_b)  # [B, 3, 1]
    col1 = make_col(a1, a0, zeros_b)  # [B, 3, 1]
    col2 = make_col(zeros_b, zeros_b, ones_b)  # [B, 3, 1]
    mat = ov_opset.concat([col0, col1, col2], axis=2).output(0)  # [B, 3, 3]

    offset_row = ov_opset.unsqueeze(b2, axes=[1]).output(0)  # [B, 1]
    offset_col = ov_opset.unsqueeze(a2, axes=[1]).output(0)  # [B, 1]
    offset_chan = ov_opset.broadcast(
        ov_opset.constant(0.0, Type.f32).output(0),
        ov_opset.reshape(
            B, ov_opset.constant([1], Type.i32).output(0), False
        ).output(0),
    ).output(0)
    offset_chan = ov_opset.unsqueeze(offset_chan, axes=[1]).output(0)  # [B, 1]
    offset_3 = ov_opset.concat(
        [offset_row, offset_col, offset_chan], axis=1
    ).output(0)  # [B, 3]

    pts_b = ov_opset.unsqueeze(pts, axes=[0]).output(0)  # [1, N, 3]
    pts_shape = ov_opset.shape_of(pts, output_type=Type.i32).output(0)  # [2]
    B_N_3 = ov_opset.concat([s1d(B), pts_shape], axis=0).output(0)  # [B, N, 3]
    pts_b = ov_opset.broadcast(pts_b, B_N_3).output(0)  # [B, N, 3]

    transformed = ov_opset.matmul(pts_b, mat, False, False).output(
        0
    )  # [B, N, 3]

    offset_b = ov_opset.unsqueeze(offset_3, axes=[1]).output(0)  # [B, 1, 3]
    transformed = ov_opset.add(transformed, offset_b).output(0)  # [B, N, 3]

    transformed_t = ov_opset.transpose(
        transformed, ov_opset.constant([0, 2, 1], Type.i32).output(0)
    ).output(0)  # [B, 3, N]
    coord_row = ov_opset.squeeze(
        ov_opset.gather(
            transformed_t,
            ov_opset.constant([0], Type.i32).output(0),
            ov_opset.constant(1, Type.i32).output(0),
        ).output(0),  # [B, 1, N]
        axes=[1],
    ).output(0)  # [B, N]
    coord_col = ov_opset.squeeze(
        ov_opset.gather(
            transformed_t,
            ov_opset.constant([1], Type.i32).output(0),
            ov_opset.constant(1, Type.i32).output(0),
        ).output(0),  # [B, 1, N]
        axes=[1],
    ).output(0)  # [B, N]
    bhwc_shape = ov_opset.concat(
        [s1d(B), s1d(H), s1d(W), s1d(C)], axis=0
    ).output(0)
    coord_row = ov_opset.reshape(coord_row, bhwc_shape, False).output(0)
    coord_col = ov_opset.reshape(coord_col, bhwc_shape, False).output(0)
    # chan coords are the same for all batch items
    chan_bhwc = ov_opset.broadcast(
        ov_opset.reshape(
            ov_opset.reshape(grid_c, neg1, False).output(0),
            ov_opset.concat(
                [
                    ov_opset.constant([1], Type.i32).output(0),
                    s1d(H),
                    s1d(W),
                    s1d(C),
                ],
                axis=0,
            ).output(0),
            False,
        ).output(0),
        bhwc_shape,
    ).output(0)
    zero_f = ov_opset.constant(0.0, Type.f32).output(0)
    one_f = ov_opset.constant(1.0, Type.f32).output(0)
    B_f = ov_opset.convert(B, Type.f32).output(0)
    r_b = ov_opset.range(zero_f, B_f, one_f, output_type=Type.f32).output(0)
    batch_bhwc = ov_opset.broadcast(
        ov_opset.reshape(
            r_b,
            ov_opset.concat(
                [s1d(B), ov_opset.constant([1, 1, 1], Type.i32).output(0)],
                axis=0,
            ).output(0),
            False,
        ).output(0),
        bhwc_shape,
    ).output(0)

    # Stack to [4, B, H, W, C] — (batch, row, col, chan)
    coords = ov_opset.concat(
        [
            ov_opset.unsqueeze(batch_bhwc, axes=[0]).output(0),
            ov_opset.unsqueeze(coord_row, axes=[0]).output(0),
            ov_opset.unsqueeze(coord_col, axes=[0]).output(0),
            ov_opset.unsqueeze(chan_bhwc, axes=[0]).output(0),
        ],
        axis=0,
    ).output(0)
    return coords


def affine_transform(
    images,
    transform,
    interpolation="bilinear",
    fill_mode="constant",
    fill_value=0,
    data_format=None,
):
    data_format = backend.standardize_data_format(data_format)
    if interpolation not in AFFINE_TRANSFORM_INTERPOLATIONS:
        raise ValueError(
            "Invalid value for argument `interpolation`. Expected of one "
            f"{set(AFFINE_TRANSFORM_INTERPOLATIONS.keys())}. Received: "
            f"interpolation={interpolation}"
        )
    if fill_mode not in AFFINE_TRANSFORM_FILL_MODES:
        raise ValueError(
            "Invalid value for argument `fill_mode`. Expected of one "
            f"{AFFINE_TRANSFORM_FILL_MODES}. Received: fill_mode={fill_mode}"
        )

    images = convert_to_tensor(images)
    transform = convert_to_tensor(transform)
    images_ov = get_ov_output(images)
    transform_ov = get_ov_output(transform)

    if len(images.shape) not in (3, 4):
        raise ValueError(
            "Invalid images rank: expected rank 3 (single image) "
            "or rank 4 (batch of images). Received input with shape: "
            f"images.shape={images.shape}"
        )
    if len(transform.shape) not in (1, 2):
        raise ValueError(
            "Invalid transform rank: expected rank 1 (single transform) "
            "or rank 2 (batch of transforms). Received input with shape: "
            f"transform.shape={transform.shape}"
        )

    ov_type = images_ov.get_element_type()
    compute_type = Type.f32

    need_squeeze = False
    if len(images.shape) == 3:
        images_ov = ov_opset.unsqueeze(images_ov, axes=[0]).output(0)
        need_squeeze = True
    if len(transform.shape) == 1:
        transform_ov = ov_opset.unsqueeze(transform_ov, axes=[0]).output(0)

    if data_format == "channels_first":
        images_ov = ov_opset.transpose(
            images_ov,
            ov_opset.constant([0, 2, 3, 1], Type.i32).output(0),
        ).output(0)

    images_ov = ov_opset.convert(images_ov, compute_type).output(0)
    transform_ov = ov_opset.convert(transform_ov, compute_type).output(0)

    coords = _ov_build_affine_coords(images_ov, transform_ov)
    affined = map_coordinates(
        OpenVINOKerasTensor(images_ov),
        OpenVINOKerasTensor(coords),
        order=AFFINE_TRANSFORM_INTERPOLATIONS[interpolation],
        fill_mode=fill_mode,
        fill_value=fill_value,
    )
    affined = get_ov_output(affined)

    if ov_type.is_integral():
        affined = ov_opset.round(affined, mode="half_to_even").output(0)
    affined = ov_opset.convert(affined, ov_type).output(0)

    if data_format == "channels_first":
        affined = ov_opset.transpose(
            affined,
            ov_opset.constant([0, 3, 1, 2], Type.i32).output(0),
        ).output(0)
    if need_squeeze:
        affined = ov_opset.squeeze(affined, axes=[0]).output(0)

    return OpenVINOKerasTensor(affined)


def compute_homography_matrix(start_points, end_points):
    start_points = convert_to_tensor(start_points, dtype="float32")
    end_points = convert_to_tensor(end_points, dtype="float32")
    sp_ov = get_ov_output(start_points)
    ep_ov = get_ov_output(end_points)

    axis0 = ov_opset.constant(0, Type.i32).output(0)
    axis1 = ov_opset.constant(1, Type.i32).output(0)

    def _split_points(pts_ov):
        corners = [
            ov_opset.squeeze(
                ov_opset.gather(
                    pts_ov,
                    ov_opset.constant([c], Type.i32).output(0),
                    axis1,
                ).output(0),
                axes=[1],
            ).output(0)
            for c in range(4)
        ]
        result = []
        for pt in corners:
            x = ov_opset.squeeze(
                ov_opset.gather(
                    pt, ov_opset.constant([0], Type.i32).output(0), axis1
                ).output(0),
                axes=[1],
            ).output(0)
            y = ov_opset.squeeze(
                ov_opset.gather(
                    pt, ov_opset.constant([1], Type.i32).output(0), axis1
                ).output(0),
                axes=[1],
            ).output(0)
            result.append((x, y))
        return result

    end_pts = _split_points(ep_ov)
    start_pts = _split_points(sp_ov)

    B_shape = ov_opset.shape_of(sp_ov, output_type=Type.i32).output(0)
    B = ov_opset.gather(
        B_shape, ov_opset.constant(0, Type.i32).output(0), axis0
    ).output(0)

    def ones():
        return ov_opset.broadcast(
            ov_opset.constant(1.0, Type.f32).output(0),
            ov_opset.reshape(
                B, ov_opset.constant([1], Type.i32).output(0), False
            ).output(0),
        ).output(0)

    def zeros():
        return ov_opset.broadcast(
            ov_opset.constant(0.0, Type.f32).output(0),
            ov_opset.reshape(
                B, ov_opset.constant([1], Type.i32).output(0), False
            ).output(0),
        ).output(0)

    def neg(x):
        return ov_opset.multiply(
            x, ov_opset.constant(-1.0, Type.f32).output(0)
        ).output(0)

    def mul(a, b):
        return ov_opset.multiply(a, b).output(0)

    def make_two_rows(ex, ey, sx, sy):
        o = ones()
        z = zeros()

        def row_from_elems(elems):
            return ov_opset.concat(
                [ov_opset.unsqueeze(e, axes=[1]).output(0) for e in elems],
                axis=1,
            ).output(0)

        row1 = row_from_elems(
            [ex, ey, o, z, z, z, neg(mul(sx, ex)), neg(mul(sx, ey))]
        )
        row2 = row_from_elems(
            [z, z, z, ex, ey, o, neg(mul(sy, ex)), neg(mul(sy, ey))]
        )
        return row1, row2

    rows = []
    rhs_cols = []
    for corner in range(4):
        ex, ey = end_pts[corner]
        sx, sy = start_pts[corner]
        r1, r2 = make_two_rows(ex, ey, sx, sy)
        rows.extend([r1, r2])
        rhs_cols.extend([sx, sy])

    coeff_mat = ov_opset.concat(
        [ov_opset.unsqueeze(r, axes=[1]).output(0) for r in rows], axis=1
    ).output(0)  # [B, 8, 8]
    rhs_2d = ov_opset.concat(
        [ov_opset.unsqueeze(c, axes=[1]).output(0) for c in rhs_cols], axis=1
    ).output(0)  # [B, 8]
    rhs = ov_opset.unsqueeze(rhs_2d, axes=[2]).output(0)  # [B, 8, 1]

    coeff_inv = ov_opset.inverse(coeff_mat, adjoint=False).output(0)
    h = ov_opset.matmul(coeff_inv, rhs, False, False).output(0)  # [B, 8, 1]
    h = ov_opset.squeeze(h, axes=[2]).output(0)  # [B, 8]
    return h


def perspective_transform(
    images,
    start_points,
    end_points,
    interpolation="bilinear",
    fill_value=0,
    data_format=None,
):
    data_format = backend.standardize_data_format(data_format)
    if interpolation not in AFFINE_TRANSFORM_INTERPOLATIONS:
        raise ValueError(
            "Invalid value for argument `interpolation`. Expected of one "
            f"{set(AFFINE_TRANSFORM_INTERPOLATIONS.keys())}. Received: "
            f"interpolation={interpolation}"
        )

    images = convert_to_tensor(images)
    start_points = convert_to_tensor(start_points)
    end_points = convert_to_tensor(end_points)
    images_ov = get_ov_output(images)
    sp_ov = get_ov_output(start_points)
    ep_ov = get_ov_output(end_points)

    if len(images.shape) not in (3, 4):
        raise ValueError(
            "Invalid images rank: expected rank 3 (single image) "
            "or rank 4 (batch of images). Received input with shape: "
            f"images.shape={images.shape}"
        )
    if len(start_points.shape) not in (2, 3) or tuple(start_points.shape)[
        -2:
    ] != (4, 2):
        raise ValueError(
            "Invalid start_points shape: expected (4,2) for a single image"
            f" or (N,4,2) for a batch. Received shape: {start_points.shape}"
        )
    if len(end_points.shape) not in (2, 3) or tuple(end_points.shape)[-2:] != (
        4,
        2,
    ):
        raise ValueError(
            "Invalid end_points shape: expected (4,2) for a single image"
            f" or (N,4,2) for a batch. Received shape: {end_points.shape}"
        )

    ov_type = images_ov.get_element_type()
    compute_type = Type.f32

    need_squeeze = False
    if len(images.shape) == 3:
        images_ov = ov_opset.unsqueeze(images_ov, axes=[0]).output(0)
        need_squeeze = True
    if len(start_points.shape) == 2:
        sp_ov = ov_opset.unsqueeze(sp_ov, axes=[0]).output(0)
    if len(end_points.shape) == 2:
        ep_ov = ov_opset.unsqueeze(ep_ov, axes=[0]).output(0)

    if data_format == "channels_first":
        images_ov = ov_opset.transpose(
            images_ov,
            ov_opset.constant([0, 2, 3, 1], Type.i32).output(0),
        ).output(0)

    images_ov = ov_opset.convert(images_ov, compute_type).output(0)

    transforms = get_ov_output(
        compute_homography_matrix(start_points, end_points)
    )

    shape_node = ov_opset.shape_of(images_ov, output_type=Type.i32).output(0)
    axis0 = ov_opset.constant(0, Type.i32).output(0)

    def dim(i):
        return ov_opset.gather(
            shape_node, ov_opset.constant(i, Type.i32).output(0), axis0
        ).output(0)

    B = dim(0)
    H = dim(1)
    W = dim(2)
    C = dim(3)

    H_f = ov_opset.convert(H, Type.f32).output(0)
    W_f = ov_opset.convert(W, Type.f32).output(0)
    zero_f = ov_opset.constant(0.0, Type.f32).output(0)
    one_f = ov_opset.constant(1.0, Type.f32).output(0)
    r_h = ov_opset.range(zero_f, H_f, one_f, output_type=Type.f32).output(
        0
    )  # [H]
    r_w = ov_opset.range(zero_f, W_f, one_f, output_type=Type.f32).output(
        0
    )  # [W]

    def p1d(scalar):
        return ov_opset.reshape(
            scalar, ov_opset.constant([1], Type.i32).output(0), False
        ).output(0)

    hw_shape = ov_opset.concat([p1d(H), p1d(W)], axis=0).output(0)
    # y: rows  [H, W]
    y = ov_opset.broadcast(
        ov_opset.reshape(
            r_h,
            ov_opset.concat(
                [p1d(H), ov_opset.constant([1], Type.i32).output(0)], axis=0
            ).output(0),
            False,
        ).output(0),
        hw_shape,
    ).output(0)
    # x: cols  [H, W]
    x = ov_opset.broadcast(
        ov_opset.reshape(
            r_w,
            ov_opset.concat(
                [ov_opset.constant([1], Type.i32).output(0), p1d(W)], axis=0
            ).output(0),
            False,
        ).output(0),
        hw_shape,
    ).output(0)

    neg1 = ov_opset.constant([-1], Type.i32).output(0)
    x_flat = ov_opset.reshape(x, neg1, False).output(0)  # [N]
    y_flat = ov_opset.reshape(y, neg1, False).output(0)  # [N]

    axis1 = ov_opset.constant(1, Type.i32).output(0)

    def h_col(i):
        return ov_opset.squeeze(
            ov_opset.gather(
                transforms, ov_opset.constant([i], Type.i32).output(0), axis1
            ).output(0),  # [B, 1]
            axes=[1],
        ).output(0)  # [B]

    a0, a1, a2 = h_col(0), h_col(1), h_col(2)
    a3, a4, a5 = h_col(3), h_col(4), h_col(5)
    a6, a7 = h_col(6), h_col(7)

    N_shape = ov_opset.shape_of(x_flat, output_type=Type.i32).output(0)
    BN_shape = ov_opset.concat(
        [
            ov_opset.reshape(
                B, ov_opset.constant([1], Type.i32).output(0), False
            ).output(0),
            N_shape,
        ],
        axis=0,
    ).output(0)

    def bcast(v):
        return ov_opset.broadcast(
            ov_opset.unsqueeze(v, axes=[1]).output(0), BN_shape
        ).output(0)

    x_bn = ov_opset.broadcast(
        ov_opset.unsqueeze(x_flat, axes=[0]).output(0), BN_shape
    ).output(0)
    y_bn = ov_opset.broadcast(
        ov_opset.unsqueeze(y_flat, axes=[0]).output(0), BN_shape
    ).output(0)

    # denom = a6*x + a7*y + 1
    one_bn = ov_opset.broadcast(
        ov_opset.constant(1.0, Type.f32).output(0), BN_shape
    ).output(0)
    denom = ov_opset.add(
        ov_opset.add(
            ov_opset.multiply(bcast(a6), x_bn).output(0),
            ov_opset.multiply(bcast(a7), y_bn).output(0),
        ).output(0),
        one_bn,
    ).output(0)

    # x_in = (a0*x + a1*y + a2) / denom
    x_in = ov_opset.divide(
        ov_opset.add(
            ov_opset.add(
                ov_opset.multiply(bcast(a0), x_bn).output(0),
                ov_opset.multiply(bcast(a1), y_bn).output(0),
            ).output(0),
            bcast(a2),
        ).output(0),
        denom,
    ).output(0)

    # y_in = (a3*x + a4*y + a5) / denom
    y_in = ov_opset.divide(
        ov_opset.add(
            ov_opset.add(
                ov_opset.multiply(bcast(a3), x_bn).output(0),
                ov_opset.multiply(bcast(a4), y_bn).output(0),
            ).output(0),
            bcast(a5),
        ).output(0),
        denom,
    ).output(0)

    bhw_shape = ov_opset.concat([p1d(B), p1d(H), p1d(W)], axis=0).output(0)
    y_in = ov_opset.reshape(y_in, bhw_shape, False).output(0)
    x_in = ov_opset.reshape(x_in, bhw_shape, False).output(0)

    C_f = ov_opset.convert(C, Type.f32).output(0)
    r_c = ov_opset.range(zero_f, C_f, one_f, output_type=Type.f32).output(
        0
    )  # [C]

    bhwc_shape = ov_opset.concat(
        [p1d(B), p1d(H), p1d(W), p1d(C)], axis=0
    ).output(0)

    y_in_bhwc = ov_opset.broadcast(
        ov_opset.unsqueeze(y_in, axes=[3]).output(0), bhwc_shape
    ).output(0)
    x_in_bhwc = ov_opset.broadcast(
        ov_opset.unsqueeze(x_in, axes=[3]).output(0), bhwc_shape
    ).output(0)
    chan_bhwc = ov_opset.broadcast(
        ov_opset.reshape(
            r_c,
            ov_opset.concat(
                [ov_opset.constant([1, 1, 1], Type.i32).output(0), p1d(C)],
                axis=0,
            ).output(0),
            False,
        ).output(0),
        bhwc_shape,
    ).output(0)
    B_f = ov_opset.convert(B, Type.f32).output(0)
    r_b = ov_opset.range(zero_f, B_f, one_f, output_type=Type.f32).output(0)
    batch_bhwc = ov_opset.broadcast(
        ov_opset.reshape(
            r_b,
            ov_opset.concat(
                [p1d(B), ov_opset.constant([1, 1, 1], Type.i32).output(0)],
                axis=0,
            ).output(0),
            False,
        ).output(0),
        bhwc_shape,
    ).output(0)

    # coords: [4, B, H, W, C] — (batch, row=y, col=x, chan)
    coords = ov_opset.concat(
        [
            ov_opset.unsqueeze(batch_bhwc, axes=[0]).output(0),
            ov_opset.unsqueeze(y_in_bhwc, axes=[0]).output(0),
            ov_opset.unsqueeze(x_in_bhwc, axes=[0]).output(0),
            ov_opset.unsqueeze(chan_bhwc, axes=[0]).output(0),
        ],
        axis=0,
    ).output(0)

    result = map_coordinates(
        OpenVINOKerasTensor(images_ov),
        OpenVINOKerasTensor(coords),
        order=AFFINE_TRANSFORM_INTERPOLATIONS[interpolation],
        fill_mode="constant",
        fill_value=fill_value,
    )
    result = get_ov_output(result)

    if ov_type.is_integral():
        result = ov_opset.round(result, mode="half_to_even").output(0)
    result = ov_opset.convert(result, ov_type).output(0)

    if data_format == "channels_first":
        result = ov_opset.transpose(
            result,
            ov_opset.constant([0, 3, 1, 2], Type.i32).output(0),
        ).output(0)
    if need_squeeze:
        result = ov_opset.squeeze(result, axes=[0]).output(0)

    return OpenVINOKerasTensor(result)


def _mirror_index_fixer(index, size_node):
    """size_node: OV i32 scalar node."""
    one_c = ov_opset.constant(1, dtype=Type.i32).output(0)
    two_c = ov_opset.constant(2, dtype=Type.i32).output(0)
    zero_c = ov_opset.constant(0, dtype=Type.i32).output(0)
    s = ov_opset.subtract(size_node, one_c).output(0)  # size - 1
    s2 = ov_opset.multiply(two_c, s).output(0)  # 2 * (size - 1)
    # Guard mod-by-zero when size == 1
    safe_s2 = ov_opset.maximum(s2, one_c).output(0)
    diff = ov_opset.subtract(
        ov_opset.mod(ov_opset.add(index, s), safe_s2), s
    ).output(0)
    # Use max(d, 0-d) instead of abs(d): abs on integers is unreliable
    # under OV's INFERENCE_PRECISION_HINT=f32 for certain graph topologies.
    neg_diff = ov_opset.subtract(zero_c, diff).output(0)
    result = ov_opset.maximum(diff, neg_diff).output(0)
    # When size == 1 all indices map to 0.
    size_is_one = ov_opset.equal(size_node, one_c).output(0)
    return ov_opset.select(size_is_one, zero_c, result).output(0)


def _reflect_index_fixer(index, size_node):
    """size_node: OV i32 scalar node."""
    # floor_divide(_mirror(2*index + 1, 2*size + 1) - 1, 2)
    two_c = ov_opset.constant(2, dtype=Type.i32).output(0)
    one_c = ov_opset.constant(1, dtype=Type.i32).output(0)
    idx2 = ov_opset.add(ov_opset.multiply(two_c, index), one_c).output(0)
    size2_plus1 = ov_opset.add(
        ov_opset.multiply(two_c, size_node), one_c
    ).output(0)
    mirrored = _mirror_index_fixer(idx2, size2_plus1)
    # integer divide by 2 (result is always >= 0, so truncated == floor)
    return ov_opset.divide(ov_opset.subtract(mirrored, one_c), two_c).output(0)


def map_coordinates(
    inputs, coordinates, order, fill_mode="constant", fill_value=0
):
    fill_modes = {"constant", "nearest", "wrap", "mirror", "reflect"}
    if fill_mode not in fill_modes:
        raise ValueError(
            "Invalid value for argument `fill_mode`. Expected one of "
            f"{fill_modes}. Received: fill_mode={fill_mode}"
        )
    if order not in (0, 1):
        raise ValueError(
            "Invalid value for argument `order`. Expected one of "
            f"[0, 1]. Received: order={order}"
        )

    inputs = convert_to_tensor(inputs)
    coordinates = convert_to_tensor(coordinates)
    inputs_ov = get_ov_output(inputs)
    coords_ov = get_ov_output(coordinates)
    input_shape = inputs_ov.get_partial_shape()
    ndim = input_shape.rank.get_length()
    coords_shape = coords_ov.get_partial_shape()
    if coords_shape.rank.is_static and coords_shape.rank.get_length() < 2:
        raise ValueError(
            "Invalid coordinates rank: expected at least rank 2."
            f" Received input with shape: {tuple(coords_shape.to_shape())}"
        )
    if coords_shape[0].is_static and coords_shape[0].get_length() != ndim:
        raise ValueError(
            "First dim of `coordinates` must be the same as the rank of "
            "`inputs`. "
            f"Received inputs with shape: {tuple(input_shape.to_shape())} and "
            f"coordinate leading dim of {coords_shape[0].get_length()}"
        )
    ov_type = inputs_ov.get_element_type()

    # Coordinates must be float for arithmetic
    coords_ov = ov_opset.convert(coords_ov, Type.f32).output(0)

    # Split coordinates into per-dimension tensors, each shape [*output_shape]
    axis_0 = ov_opset.constant(0, dtype=Type.i32).output(0)
    coord_list = [
        ov_opset.gather(
            coords_ov,
            ov_opset.constant(i, dtype=Type.i32).output(0),
            axis_0,
        ).output(0)
        for i in range(ndim)
    ]

    input_shape_node = ov_opset.shape_of(
        inputs_ov, output_type=Type.i32
    ).output(0)
    size_nodes = [
        ov_opset.gather(
            input_shape_node,
            ov_opset.constant(i, dtype=Type.i32).output(0),
            axis_0,
        ).output(0)
        for i in range(ndim)
    ]

    def fix_index(index, size_node):
        if fill_mode in ("constant", "nearest"):
            zero = ov_opset.constant(0, dtype=Type.i32).output(0)
            size_m1 = ov_opset.subtract(
                size_node, ov_opset.constant(1, dtype=Type.i32)
            ).output(0)
            return ov_opset.minimum(
                ov_opset.maximum(index, zero), size_m1
            ).output(0)
        elif fill_mode == "wrap":
            return ov_opset.floor_mod(index, size_node).output(0)
        elif fill_mode == "mirror":
            return _mirror_index_fixer(index, size_node)
        else:  # reflect
            return _reflect_index_fixer(index, size_node)

    def is_valid(index, size_node):
        if fill_mode != "constant":
            return None
        return ov_opset.logical_and(
            ov_opset.greater_equal(index, ov_opset.constant(0, dtype=Type.i32)),
            ov_opset.less(index, size_node),
        ).output(0)

    # Build per-dimension interpolation nodes
    interp_nodes_per_dim = []
    for i, coord in enumerate(coord_list):
        size_node = size_nodes[i]
        if order == 0:
            idx = ov_opset.convert(
                ov_opset.round(coord, mode="half_to_even"), Type.i32
            ).output(0)
            interp_nodes_per_dim.append(
                [(fix_index(idx, size_node), is_valid(idx, size_node), None)]
            )
        else:
            lower_f = ov_opset.floor(coord).output(0)
            upper_w = ov_opset.subtract(coord, lower_f).output(0)
            lower_w = ov_opset.subtract(
                ov_opset.constant(1.0, dtype=Type.f32), upper_w
            ).output(0)
            lower_idx = ov_opset.convert(lower_f, Type.i32).output(0)
            upper_idx = ov_opset.add(
                lower_idx, ov_opset.constant(1, dtype=Type.i32)
            ).output(0)
            interp_nodes_per_dim.append(
                [
                    (
                        fix_index(lower_idx, size_node),
                        is_valid(lower_idx, size_node),
                        lower_w,
                    ),
                    (
                        fix_index(upper_idx, size_node),
                        is_valid(upper_idx, size_node),
                        upper_w,
                    ),
                ]
            )

    fill_const = ov_opset.convert(
        ov_opset.constant(float(fill_value), dtype=Type.f32), ov_type
    ).output(0)

    output = None
    for items in itertools.product(*interp_nodes_per_dim):
        indices, validities, weights = zip(*items)

        # Stack indices: [*output_shape, ndim] for gather_nd
        stacked = ov_opset.concat(
            [ov_opset.unsqueeze(idx, axes=[-1]).output(0) for idx in indices],
            axis=-1,
        ).output(0)
        gathered = ov_opset.gather_nd(inputs_ov, stacked).output(0)

        if fill_mode == "constant":
            valid = validities[0]
            for v in validities[1:]:
                valid = ov_opset.logical_and(valid, v).output(0)
            gathered = ov_opset.select(valid, gathered, fill_const).output(0)

        if any(w is not None for w in weights):
            result_type = ov_type if not ov_type.is_integral() else Type.f32
            contribution = ov_opset.convert(gathered, result_type).output(0)
            for w in weights:
                if w is not None:
                    contribution = ov_opset.multiply(
                        contribution,
                        ov_opset.convert(w, result_type).output(0),
                    ).output(0)
        else:
            contribution = gathered

        output = (
            contribution
            if output is None
            else ov_opset.add(output, contribution).output(0)
        )

    if ov_type.is_integral() and order == 1:
        output = ov_opset.convert(
            ov_opset.round(
                ov_opset.convert(output, Type.f32).output(0),
                mode="half_to_even",
            ).output(0),
            ov_type,
        ).output(0)

    return OpenVINOKerasTensor(output)


def gaussian_blur(
    images, kernel_size=(3, 3), sigma=(1.0, 1.0), data_format=None
):
    def _create_gaussian_kernel(kernel_size, sigma):
        # Always build the kernel in f32 for numerical stability and
        # compatibility (bfloat16 / f16 are not fully supported by all ops).
        def _get_gaussian_kernel1d(size, sigma):
            x = ov_opset.subtract(
                ov_opset.range(0, size, 1, output_type=Type.f32).output(0),
                ov_opset.constant((size - 1) / 2.0, Type.f32).output(0),
            ).output(0)

            sigma_const = ov_opset.constant(float(sigma), Type.f32).output(0)
            exponent = ov_opset.divide(x, sigma_const).output(0)
            exponent = ov_opset.multiply(exponent, exponent).output(0)
            exponent = ov_opset.multiply(
                exponent, ov_opset.constant(-0.5, Type.f32).output(0)
            ).output(0)
            kernel1d = ov_opset.exp(exponent).output(0)
            kernel1d_sum = ov_opset.reduce_sum(
                kernel1d, reduction_axes=0, keep_dims=False
            ).output(0)

            return ov_opset.divide(kernel1d, kernel1d_sum).output(0)

        def _get_gaussian_kernel2d(size, sigma):
            kernel1d_x = _get_gaussian_kernel1d(size[0], sigma[0])
            kernel1d_y = _get_gaussian_kernel1d(size[1], sigma[1])

            # kernel1d_x has kH elements -> row vector [1, kH]
            kernel1d_x = ov_opset.reshape(
                kernel1d_x,
                ov_opset.constant([1, int(size[0])], Type.i32).output(0),
                False,
            ).output(0)

            # kernel1d_y has kW elements -> column vector [kW, 1]
            kernel1d_y = ov_opset.reshape(
                kernel1d_y,
                ov_opset.constant([int(size[1]), 1], Type.i32).output(0),
                False,
            ).output(0)
            return ov_opset.multiply(kernel1d_y, kernel1d_x).output(0)

        return _get_gaussian_kernel2d(kernel_size, sigma)

    data_format = backend.standardize_data_format(data_format)
    images = convert_to_tensor(images)
    input_shape = images.shape
    ov_type = get_ov_output(images).get_element_type()

    if len(input_shape) not in (3, 4):
        raise ValueError(
            "Invalid images rank: expected rank 3 (single image) "
            "or rank 4 (batch of images). Received input with shape: "
            f"images.shape={input_shape}"
        )

    # bfloat16 is not supported by all OV ops used here; promote to f32.
    # f16 constants in range() would mismatch with the f32 arithmetic
    # constants in the kernel builder, so promote f16 to f32 as well.
    compute_type = Type.f32 if ov_type in (Type.bf16, Type.f16) else ov_type
    images = get_ov_output(images)
    if compute_type != ov_type:
        images = ov_opset.convert(images, compute_type).output(0)

    need_squeeze = False
    if len(input_shape) == 3:
        images = ov_opset.unsqueeze(images, axes=[0]).output(0)
        need_squeeze = True

    if data_format == "channels_last":
        images = ov_opset.transpose(
            images,
            ov_opset.constant([0, 3, 1, 2], Type.i32).output(0),
        ).output(0)

    num_channels = ov_opset.gather(
        ov_opset.shape_of(images, Type.i32).output(0),
        ov_opset.constant([1], Type.i32).output(0),
        ov_opset.constant(0, Type.i32).output(0),
    ).output(0)
    # Kernel is always built in f32; convert to compute_type before conv.
    kernel = _create_gaussian_kernel(kernel_size, sigma)
    kernel = ov_opset.convert(kernel, compute_type).output(0)

    kernel = ov_opset.reshape(
        kernel,
        ov_opset.constant(
            [1, 1, kernel_size[0], kernel_size[1]], Type.i32
        ).output(0),
        False,
    ).output(0)

    target_shape = ov_opset.concat(
        [
            num_channels,
            ov_opset.constant(
                [1, 1, kernel_size[0], kernel_size[1]], Type.i32
            ).output(0),
        ],
        axis=0,
    ).output(0)

    kernel = ov_opset.broadcast(kernel, target_shape).output(0)

    blurred_images = ov_opset.group_convolution(
        images,
        kernel,
        [1, 1],
        [(kernel_size[0] - 1) // 2, (kernel_size[1] - 1) // 2],
        [
            kernel_size[0] - 1 - (kernel_size[0] - 1) // 2,
            kernel_size[1] - 1 - (kernel_size[1] - 1) // 2,
        ],
        [1, 1],
    ).output(0)

    # Cast back to the original dtype if we promoted for computation.
    if compute_type != ov_type:
        blurred_images = ov_opset.convert(blurred_images, ov_type).output(0)

    if data_format == "channels_last":
        blurred_images = ov_opset.transpose(
            blurred_images,
            ov_opset.constant([0, 2, 3, 1], Type.i32).output(0),
        ).output(0)

    if need_squeeze:
        blurred_images = ov_opset.squeeze(blurred_images, axes=[0]).output(0)

    return OpenVINOKerasTensor(blurred_images)


def elastic_transform(
    images,
    alpha=20.0,
    sigma=5.0,
    interpolation="bilinear",
    fill_mode="reflect",
    fill_value=0.0,
    seed=None,
    data_format=None,
):
    data_format = backend.standardize_data_format(data_format)
    if interpolation not in AFFINE_TRANSFORM_INTERPOLATIONS:
        raise ValueError(
            "Invalid value for argument `interpolation`. Expected of one "
            f"{set(AFFINE_TRANSFORM_INTERPOLATIONS.keys())}. Received: "
            f"interpolation={interpolation}"
        )
    if fill_mode not in AFFINE_TRANSFORM_FILL_MODES:
        raise ValueError(
            "Invalid value for argument `fill_mode`. Expected of one "
            f"{AFFINE_TRANSFORM_FILL_MODES}. Received: fill_mode={fill_mode}"
        )
    if len(images.shape) not in (3, 4):
        raise ValueError(
            "Invalid images rank: expected rank 3 (single image) "
            "or rank 4 (batch of images). Received input with shape: "
            f"images.shape={images.shape}"
        )

    images = convert_to_tensor(images)
    images_ov = get_ov_output(images)
    ov_type = images_ov.get_element_type()
    compute_type = Type.f32

    need_squeeze = False
    if len(images.shape) == 3:
        images_ov = ov_opset.unsqueeze(images_ov, axes=[0]).output(0)
        need_squeeze = True

    if data_format == "channels_last":
        images_ov_cf = ov_opset.transpose(
            images_ov,
            ov_opset.constant([0, 3, 1, 2], Type.i32).output(0),
        ).output(0)
    else:
        images_ov_cf = images_ov

    images_ov_cf = ov_opset.convert(images_ov_cf, compute_type).output(0)

    shape_node = ov_opset.shape_of(images_ov_cf, output_type=Type.i32).output(0)
    axis0 = ov_opset.constant(0, Type.i32).output(0)

    def dim(i):
        return ov_opset.gather(
            shape_node, ov_opset.constant(i, Type.i32).output(0), axis0
        ).output(0)

    B = dim(0)
    C = dim(1)
    H = dim(2)
    W = dim(3)

    sigma_val = float(sigma)
    alpha_val = float(alpha)
    kernel_size_1d = int(6 * sigma_val) | 1

    # OV random ops require static seed attributes, so symbolic seeds must be
    # materialized via convert_to_numpy. This is an unavoidable sync point given
    # the OV backend's stateless random design.
    seed_val = draw_seed(seed)
    if isinstance(seed_val, OpenVINOKerasTensor):
        s = convert_to_numpy(seed_val)
    else:
        s = seed_val.data
    seed1 = max(1, int(s[0]) & 0x7FFFFFFF)
    seed2 = max(1, int(s[1]) & 0x7FFFFFFF) if len(s) > 1 else 1

    def to_1d(scalar):
        return ov_opset.reshape(
            scalar, ov_opset.constant([1], Type.i32).output(0), False
        ).output(0)

    bhw_shape = ov_opset.concat(
        [to_1d(B), to_1d(H), to_1d(W)],
        axis=0,
    ).output(0)

    dx = _random_normal(bhw_shape, Type.f32, seed1, seed2)  # [B, H, W]
    dy = _random_normal(bhw_shape, Type.f32, seed1 + 1, seed2)  # [B, H, W]

    # Scale by sigma before gaussian blur
    sigma_const = ov_opset.constant(sigma_val, Type.f32).output(0)
    dx = ov_opset.multiply(dx, sigma_const).output(0)
    dy = ov_opset.multiply(dy, sigma_const).output(0)

    # Apply gaussian blur to smooth the displacement fields
    # Add channel dim: [B, 1, H, W] for channels_first gaussian_blur
    dx_4d = ov_opset.unsqueeze(dx, axes=[1]).output(0)
    dy_4d = ov_opset.unsqueeze(dy, axes=[1]).output(0)

    dx_blurred = gaussian_blur(
        OpenVINOKerasTensor(dx_4d),
        kernel_size=(kernel_size_1d, kernel_size_1d),
        sigma=(sigma_val, sigma_val),
        data_format="channels_first",
    )
    dy_blurred = gaussian_blur(
        OpenVINOKerasTensor(dy_4d),
        kernel_size=(kernel_size_1d, kernel_size_1d),
        sigma=(sigma_val, sigma_val),
        data_format="channels_first",
    )
    dx_blurred = ov_opset.squeeze(get_ov_output(dx_blurred), axes=[1]).output(
        0
    )  # [B, H, W]
    dy_blurred = ov_opset.squeeze(get_ov_output(dy_blurred), axes=[1]).output(
        0
    )  # [B, H, W]

    H_f = ov_opset.convert(H, Type.f32).output(0)
    W_f = ov_opset.convert(W, Type.f32).output(0)
    zero_f = ov_opset.constant(0.0, Type.f32).output(0)
    one_f = ov_opset.constant(1.0, Type.f32).output(0)
    r_h = ov_opset.range(zero_f, H_f, one_f, output_type=Type.f32).output(0)
    r_w = ov_opset.range(zero_f, W_f, one_f, output_type=Type.f32).output(0)
    hw_shape = ov_opset.concat([to_1d(H), to_1d(W)], axis=0).output(0)

    y_base = ov_opset.broadcast(
        ov_opset.reshape(
            r_h,
            ov_opset.concat(
                [to_1d(H), ov_opset.constant([1], Type.i32).output(0)], axis=0
            ).output(0),
            False,
        ).output(0),
        hw_shape,
    ).output(0)  # [H, W]
    x_base = ov_opset.broadcast(
        ov_opset.reshape(
            r_w,
            ov_opset.concat(
                [ov_opset.constant([1], Type.i32).output(0), to_1d(W)], axis=0
            ).output(0),
            False,
        ).output(0),
        hw_shape,
    ).output(0)  # [H, W]

    bhw_bcast = ov_opset.concat([to_1d(B), to_1d(H), to_1d(W)], axis=0).output(
        0
    )
    y_base_b = ov_opset.broadcast(
        ov_opset.unsqueeze(y_base, axes=[0]).output(0), bhw_bcast
    ).output(0)
    x_base_b = ov_opset.broadcast(
        ov_opset.unsqueeze(x_base, axes=[0]).output(0), bhw_bcast
    ).output(0)

    alpha_const = ov_opset.constant(alpha_val, Type.f32).output(0)
    distorted_x = ov_opset.add(
        x_base_b, ov_opset.multiply(alpha_const, dx_blurred).output(0)
    ).output(0)
    distorted_y = ov_opset.add(
        y_base_b, ov_opset.multiply(alpha_const, dy_blurred).output(0)
    ).output(0)

    # Build coords [3, B, H, W, C] — (row=y, col=x, chan) per output pixel
    C_f = ov_opset.convert(C, Type.f32).output(0)
    r_c = ov_opset.range(zero_f, C_f, one_f, output_type=Type.f32).output(0)

    bhwc_shape = ov_opset.concat(
        [to_1d(B), to_1d(H), to_1d(W), to_1d(C)], axis=0
    ).output(0)

    y_bhwc = ov_opset.broadcast(
        ov_opset.unsqueeze(distorted_y, axes=[3]).output(0), bhwc_shape
    ).output(0)
    x_bhwc = ov_opset.broadcast(
        ov_opset.unsqueeze(distorted_x, axes=[3]).output(0), bhwc_shape
    ).output(0)
    chan_bhwc = ov_opset.broadcast(
        ov_opset.reshape(
            r_c,
            ov_opset.concat(
                [ov_opset.constant([1, 1, 1], Type.i32).output(0), to_1d(C)],
                axis=0,
            ).output(0),
            False,
        ).output(0),
        bhwc_shape,
    ).output(0)

    B_f = ov_opset.convert(B, Type.f32).output(0)
    r_b = ov_opset.range(zero_f, B_f, one_f, output_type=Type.f32).output(0)
    batch_bhwc = ov_opset.broadcast(
        ov_opset.reshape(
            r_b,
            ov_opset.concat(
                [to_1d(B), ov_opset.constant([1, 1, 1], Type.i32).output(0)],
                axis=0,
            ).output(0),
            False,
        ).output(0),
        bhwc_shape,
    ).output(0)

    # coords: [4, B, H, W, C] — (batch, row=y, col=x, chan)
    coords = ov_opset.concat(
        [
            ov_opset.unsqueeze(batch_bhwc, axes=[0]).output(0),
            ov_opset.unsqueeze(y_bhwc, axes=[0]).output(0),
            ov_opset.unsqueeze(x_bhwc, axes=[0]).output(0),
            ov_opset.unsqueeze(chan_bhwc, axes=[0]).output(0),
        ],
        axis=0,
    ).output(0)  # [4, B, H, W, C]

    # images_ov_cf is [B, C, H, W] but map_coordinates needs [B, H, W, C]
    images_bhwc = ov_opset.transpose(
        images_ov_cf,
        ov_opset.constant([0, 2, 3, 1], Type.i32).output(0),
    ).output(0)

    result = map_coordinates(
        OpenVINOKerasTensor(images_bhwc),
        OpenVINOKerasTensor(coords),
        order=AFFINE_TRANSFORM_INTERPOLATIONS[interpolation],
        fill_mode=fill_mode,
        fill_value=fill_value,
    )
    result = get_ov_output(result)

    if ov_type.is_integral():
        result = ov_opset.round(result, mode="half_to_even").output(0)
    result = ov_opset.convert(result, ov_type).output(0)

    if data_format == "channels_first":
        result = ov_opset.transpose(
            result,
            ov_opset.constant([0, 3, 1, 2], Type.i32).output(0),
        ).output(0)
    if need_squeeze:
        result = ov_opset.squeeze(result, axes=[0]).output(0)

    return OpenVINOKerasTensor(result)


def _ov_fill_triangle_kernel(x):
    t = x.get_element_type()
    # x is always non-negative; clamp(1 - x, 0, 1) == max(0, 1 - |x|)
    return ov_opset.clamp(
        ov_opset.subtract(ov_opset.constant(1.0, dtype=t), x), 0.0, 1.0
    ).output(0)


def _ov_fill_keys_cubic_kernel(x):
    t = x.get_element_type()
    c = lambda v: ov_opset.constant(v, dtype=t)
    # out = ((1.5x - 2.5)x)x + 1.0
    out = ov_opset.add(
        ov_opset.multiply(
            ov_opset.multiply(
                ov_opset.subtract(ov_opset.multiply(c(1.5), x), c(2.5)), x
            ),
            x,
        ),
        c(1.0),
    ).output(0)
    # out2 = ((-0.5x + 2.5)x - 4.0)x + 2.0
    out2 = ov_opset.add(
        ov_opset.multiply(
            ov_opset.subtract(
                ov_opset.multiply(
                    ov_opset.add(ov_opset.multiply(c(-0.5), x), c(2.5)), x
                ),
                c(4.0),
            ),
            x,
        ),
        c(2.0),
    ).output(0)
    out = ov_opset.select(ov_opset.greater_equal(x, c(1.0)), out2, out).output(
        0
    )
    return ov_opset.select(
        ov_opset.greater_equal(x, c(2.0)), c(0.0), out
    ).output(0)


def _ov_fill_lanczos_kernel(radius, x):
    t = x.get_element_type()
    c = lambda v: ov_opset.constant(v, dtype=t)
    pi_x = ov_opset.multiply(c(float(np.pi)), x).output(0)
    y = ov_opset.multiply(
        c(float(radius)),
        ov_opset.multiply(
            ov_opset.sin(pi_x),
            ov_opset.sin(ov_opset.divide(pi_x, c(float(radius)))),
        ),
    ).output(0)
    safe_denom = ov_opset.select(
        ov_opset.equal(x, c(0.0)),
        c(1.0),
        ov_opset.multiply(c(float(np.pi**2)), ov_opset.multiply(x, x)),
    ).output(0)
    out = ov_opset.select(
        ov_opset.greater(x, c(1e-3)), ov_opset.divide(y, safe_denom), c(1.0)
    ).output(0)
    return ov_opset.select(
        ov_opset.greater(x, c(float(radius))), c(0.0), out
    ).output(0)


_ov_kernels = {
    "linear": _ov_fill_triangle_kernel,
    "cubic": _ov_fill_keys_cubic_kernel,
    "lanczos3": lambda x: _ov_fill_lanczos_kernel(3.0, x),
    "lanczos5": lambda x: _ov_fill_lanczos_kernel(5.0, x),
}


def _ov_compute_weight_mat(
    input_size_node,
    output_size,
    scale_i,
    translation_i,
    kernel,
    antialias,
    ov_type,
):
    """Build the [m, n] resampling weight matrix.

    input_size_node: OV i32 scalar node — supports dynamic spatial dimensions.
    output_size: Python int — always statically known from the output_shape arg.
    """
    one = ov_opset.constant(1.0, dtype=ov_type).output(0)
    half = ov_opset.constant(0.5, dtype=ov_type).output(0)
    inv_scale = ov_opset.divide(one, scale_i).output(0)
    if antialias:
        kernel_scale = ov_opset.maximum(inv_scale, one).output(0)
    else:
        kernel_scale = one
    # sample_f = (arange(output_size) + 0.5) * inv_scale
    #            - translation_i * inv_scale - 0.5
    arange_n = ov_opset.constant(
        np.arange(output_size, dtype=np.float32), dtype=ov_type
    ).output(0)
    sample_f = ov_opset.subtract(
        ov_opset.subtract(
            ov_opset.multiply(ov_opset.add(arange_n, half), inv_scale),
            ov_opset.multiply(translation_i, inv_scale),
        ),
        half,
    ).output(0)
    input_size_f = ov_opset.convert(input_size_node, ov_type).output(0)
    arange_m = ov_opset.range(
        ov_opset.constant(0.0, dtype=ov_type).output(0),
        input_size_f,
        ov_opset.constant(1.0, dtype=ov_type).output(0),
        output_type=ov_type,
    ).output(0)
    sample_f_2d = ov_opset.unsqueeze(sample_f, axes=[0]).output(0)  # [1, n]
    arange_m_2d = ov_opset.unsqueeze(arange_m, axes=[1]).output(0)  # [m, 1]
    x = ov_opset.divide(
        ov_opset.abs(ov_opset.subtract(sample_f_2d, arange_m_2d)),
        kernel_scale,
    ).output(0)  # [m, n]
    weights = kernel(x)  # [m, n]
    axes_0_const = ov_opset.constant([0], Type.i32).output(0)
    total_weight_sum = ov_opset.reduce_sum(
        weights, axes_0_const, keep_dims=True
    ).output(0)  # [1, n]
    eps_val = 1000.0 * float(np.finfo(np.float32).eps)
    safe_denom = ov_opset.select(
        ov_opset.equal(total_weight_sum, ov_opset.constant(0.0, dtype=ov_type)),
        one,
        total_weight_sum,
    ).output(0)
    weights = ov_opset.select(
        ov_opset.greater(
            ov_opset.abs(total_weight_sum),
            ov_opset.constant(eps_val, dtype=ov_type),
        ),
        ov_opset.divide(weights, safe_denom),
        ov_opset.constant(0.0, dtype=ov_type),
    ).output(0)
    # Mask out-of-bounds sample positions
    upper_bound = ov_opset.subtract(
        input_size_f, ov_opset.constant(0.5, dtype=ov_type)
    ).output(0)
    in_bounds = ov_opset.logical_and(
        ov_opset.greater_equal(
            sample_f, ov_opset.constant(-0.5, dtype=ov_type)
        ),
        ov_opset.less_equal(sample_f, upper_bound),
    ).output(0)  # [n]
    in_bounds_2d = ov_opset.unsqueeze(in_bounds, axes=[0]).output(0)  # [1, n]
    return ov_opset.select(
        in_bounds_2d, weights, ov_opset.constant(0.0, dtype=ov_type)
    ).output(0)  # [m, n]


def _ov_scale_and_translate(
    x, output_shape, spatial_dims, scale_ov, translation_ov, kernel, antialias
):
    input_shape = x.get_partial_shape()
    ndim = input_shape.rank.get_length()
    x_ov_type = x.get_element_type()
    weight_ov_type = scale_ov.get_element_type()

    if len(spatial_dims) == 0:
        return OpenVINOKerasTensor(x)

    use_rounding = x_ov_type.is_integral()
    if use_rounding:
        output = ov_opset.convert(x, Type.f32).output(0)
    else:
        output = x

    axis_0 = ov_opset.constant(0, dtype=Type.i32).output(0)
    for i, d in enumerate(spatial_dims):
        d = d % ndim
        output_shape_node = ov_opset.shape_of(
            output, output_type=Type.i32
        ).output(0)
        m_node = ov_opset.gather(
            output_shape_node,
            ov_opset.constant(d, dtype=Type.i32).output(0),
            axis_0,
        ).output(0)
        n = output_shape[d]

        idx_i = ov_opset.constant(i, dtype=Type.i32).output(0)
        scale_i = ov_opset.gather(scale_ov, idx_i, axis_0).output(0)
        translation_i = ov_opset.gather(translation_ov, idx_i, axis_0).output(0)

        w = _ov_compute_weight_mat(
            m_node, n, scale_i, translation_i, kernel, antialias, weight_ov_type
        )
        w = ov_opset.convert(w, output.get_element_type()).output(0)

        # tensordot(output, w, axes=((d,), (0,))) then moveaxis(-1, d):
        # move axis d to last, matmul with w, move result back to d
        perm = list(range(ndim))
        perm.remove(d)
        perm.append(d)
        perm_const = ov_opset.constant(perm, dtype=Type.i32).output(0)
        output = ov_opset.transpose(output, perm_const).output(0)
        output = ov_opset.matmul(output, w, False, False).output(0)
        inv_perm = list(range(d)) + [ndim - 1] + list(range(d, ndim - 1))
        inv_perm_const = ov_opset.constant(inv_perm, dtype=Type.i32).output(0)
        output = ov_opset.transpose(output, inv_perm_const).output(0)

    if use_rounding:
        dtype_min = float(DTYPES_MIN[x_ov_type])
        dtype_max = float(DTYPES_MAX[x_ov_type])
        output = ov_opset.round(output, mode="half_to_even").output(0)
        output = ov_opset.clamp(output, dtype_min, dtype_max).output(0)
        output = ov_opset.convert(output, x_ov_type).output(0)

    return OpenVINOKerasTensor(output)


def scale_and_translate(
    images,
    output_shape,
    scale,
    translation,
    spatial_dims,
    method,
    antialias=True,
):
    if method not in SCALE_AND_TRANSLATE_METHODS:
        raise ValueError(
            "Invalid value for argument `method`. Expected of one "
            f"{SCALE_AND_TRANSLATE_METHODS}. Received: method={method}"
        )
    if method in ("linear", "bilinear", "trilinear", "triangle"):
        method = "linear"
    elif method in ("cubic", "bicubic", "tricubic"):
        method = "cubic"

    images = convert_to_tensor(images)
    scale = convert_to_tensor(scale)
    translation = convert_to_tensor(translation)
    dtype = backend.result_type(scale.dtype, translation.dtype)
    scale = cast(scale, dtype)
    translation = cast(translation, dtype)
    kernel = _ov_kernels[method]
    return _ov_scale_and_translate(
        get_ov_output(images),
        output_shape,
        spatial_dims,
        get_ov_output(scale),
        get_ov_output(translation),
        kernel,
        antialias,
    )


def sobel_edges(images, data_format=None):
    raise NotImplementedError(
        "`sobel_edges` is not supported with openvino backend"
    )



"""PyTorch ExportedProgram export utilities."""

import inspect
import warnings

from keras.src import tree
from keras.src.export.export_utils import convert_spec_to_tensor
from keras.src.export.export_utils import get_input_signature
from keras.src.utils import io_utils


def export_torch(
    model,
    filepath,
    input_signature=None,
    verbose=None,
    **kwargs,
):
    """Export the model as a PyTorch ExportedProgram (`.pt2`) artifact.

    Uses `torch.export.export` to capture the model's computation graph
    in an Ahead-of-Time (AOT) fashion. The resulting artifact contains
    only ATen-level operations and can be loaded via `torch.export.load`,
    run with `ExportedProgram.module()`, or compiled further with
    `torch.compile`.

    Args:
        model: The Keras model to export.
        filepath: `str` or `pathlib.Path` object. Path to save the
            exported artifact. Must end with `.pt2`.
        input_signature: Optional input signature. If `None`, inferred
            from the model's built input spec.
        verbose: `bool`. Whether to print a message after export.
            Defaults to `True`.
        **kwargs: Additional keyword arguments forwarded to
            `torch.export.export`, including `strict`, `dynamic_shapes`,
            `prefer_deferred_runtime_asserts_over_guards`, and
            `preserve_module_call_signature`. For a single-input model,
            `dynamic_shapes` can be a dict such as
            `{0: torch.export.Dim("batch", min=1, max=128)}`.
            Note: when `dynamic_shapes` is provided, the sample input used
            for tracing uses a batch size of 2 instead of 1. This avoids a
            PyTorch limitation where dimensions of size 1 are specialized
            to constants during export.

    Example:

    ```python
    model.export("path/to/model.pt2", format="torch")

    import torch
    loaded_program = torch.export.load("path/to/model.pt2")
    output = loaded_program.module()(torch.randn(1, 10))
    ```
    """
    import torch

    filepath = str(filepath)
    if not filepath.endswith(".pt2"):
        raise ValueError(
            "The PyTorch export requires the filepath to end with "
            f"'.pt2'. Got: {filepath}"
        )

    export_kwargs = _get_export_kwargs(kwargs)
    dynamic_shapes = export_kwargs.get("dynamic_shapes")

    if input_signature is None:
        input_signature = get_input_signature(model)

    # PyTorch limitation: torch.export specializes dimensions of size 1 to
    # constants during tracing. If a dynamic dim (e.g., batch) has a concrete
    # sample value of 1, export fails with "specialized it to be a constant".
    # Using a sample value of 2 (the smallest integer > 1) avoids this.
    # See: https://github.com/pytorch/pytorch/issues/176349
    replace_none_number = 2 if dynamic_shapes is not None else 1
    sample_inputs = tree.map_structure(
        lambda x: convert_spec_to_tensor(
            x, replace_none_number=replace_none_number
        ),
        input_signature,
    )
    sample_inputs = tuple(sample_inputs)
    if dynamic_shapes is not None:
        export_kwargs["dynamic_shapes"] = _normalize_dynamic_shapes(
            dynamic_shapes, sample_inputs
        )

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=r".*not properly registered as a submodule.*",
        )
        try:
            exported_program = torch.export.export(
                model,
                sample_inputs,
                **export_kwargs,
            )
        except Exception as e:
            raise RuntimeError(
                "Failed to export model to PyTorch format. "
                "Common causes: unsupported operations, data-dependent "
                "control flow, or shape constraints not satisfied. "
                f"Original error: {e}"
            ) from e

    torch.export.save(exported_program, filepath)

    actual_verbose = verbose if verbose is not None else True
    if actual_verbose:
        io_utils.print_msg(f"Saved PyTorch ExportedProgram at '{filepath}'.")
    return filepath


def _get_export_kwargs(user_kwargs):
    import torch

    supported_arg_names = {
        name
        for name in inspect.signature(torch.export.export).parameters
        if name not in {"mod", "args", "kwargs"}
    }
    unknown_arg_names = sorted(set(user_kwargs).difference(supported_arg_names))
    if unknown_arg_names:
        supported_args = ", ".join(sorted(supported_arg_names))
        unknown_args = ", ".join(unknown_arg_names)
        raise ValueError(
            'Unsupported arguments for `format="torch"`: '
            f"{unknown_args}. Supported arguments are: {supported_args}."
        )

    export_kwargs = {
        name: value
        for name, value in user_kwargs.items()
        if name in supported_arg_names
    }
    # Default to non-strict mode for broader Keras model compatibility.
    # Strict mode requires all operations to be fully ATen-traceable, which
    # is too restrictive for Keras models that use Python-level structure.
    export_kwargs.setdefault("strict", False)
    return export_kwargs


def _normalize_dynamic_shapes(dynamic_shapes, sample_inputs):
    if isinstance(dynamic_shapes, dict):
        if "args" in dynamic_shapes or "kwargs" in dynamic_shapes:
            return dynamic_shapes
        # Keras models use `forward(*args, **kwargs)`. torch.export treats
        # the entire `args` tuple as a single positional argument when the
        # forward signature is `*args`. Therefore dynamic_shapes must be a
        # list/tuple of length 1, where the single element is the spec for
        # the args tuple. For a single-input model, args = (tensor,), so
        # the spec is (dynamic_shapes,).
        if len(sample_inputs) == 1:
            return [(dynamic_shapes,)]
        return dynamic_shapes

    if isinstance(dynamic_shapes, (list, tuple)):
        # Already wrapped, e.g. [({0: Dim},)] or [([spec1, spec2],)].
        if len(dynamic_shapes) == 1 and isinstance(
            dynamic_shapes[0], (list, tuple)
        ):
            return dynamic_shapes
        # Bare list/tuple matching sample_inputs length (always 1 for
        # Keras). Wrap it in the list required by torch.export's *args
        # handling.
        if len(dynamic_shapes) == len(sample_inputs):
            return [tuple(dynamic_shapes)]

    return dynamic_shapes



import hashlib
import inspect
import os
import re
import shutil
import sys
import tarfile
import tempfile
import urllib
import urllib.error
import urllib.parse
import warnings
import zipfile
from urllib.request import urlretrieve

from keras.src.api_export import keras_export
from keras.src.backend import config
from keras.src.utils import io_utils
from keras.src.utils.module_utils import gfile
from keras.src.utils.progbar import Progbar


def path_to_string(path):
    """Convert `PathLike` objects to their string representation.

    If given a non-string typed path object, converts it to its string
    representation.

    If the object passed to `path` is not among the above, then it is
    returned unchanged. This allows e.g. passthrough of file objects
    through this function.

    Args:
        path: `PathLike` object that represents a path

    Returns:
        A string representation of the path argument, if Python support exists.
    """
    if isinstance(path, os.PathLike):
        return os.fspath(path)
    return path


def resolve_path(path):
    return os.path.realpath(path)


def resolve_sub_path(base_dir, relative_path):
    """Verify that a relative path stays within the base directory.

    Args:
        base_dir: The base directory to check against, must be resolved via
            `resolve_path`.
        relative_path: The relative path to check.

    Returns:
        The resolved path if it is within the base directory, None otherwise.
    """
    path = resolve_path(os.path.join(base_dir, relative_path))
    try:
        if os.path.commonpath([path, base_dir]) == base_dir:
            return path
        return None
    except ValueError:
        return None


def is_link_in_dir(info, base_dir):
    if info.islnk():
        # Hard links resolve relative to the root. Verify both the link
        # location and target location.
        return (
            resolve_sub_path(base_dir, info.name) is not None
            and resolve_sub_path(base_dir, info.linkname) is not None
        )

    # Symlinks resolve relative to the directory of their destination.
    destination = resolve_sub_path(base_dir, info.name)
    if destination is None:
        return False
    link = resolve_path(
        os.path.join(os.path.dirname(destination), info.linkname)
    )
    try:
        if os.path.commonpath([link, base_dir]) == base_dir:
            return True
        return False
    except ValueError:
        return False


def filter_safe_zipinfos(members, base_dir):
    base_dir = resolve_path(base_dir)
    for finfo in members:
        valid_path = False
        if resolve_sub_path(base_dir, finfo.filename) is not None:
            valid_path = True
            yield finfo
        if not valid_path:
            warnings.warn(
                "Skipping invalid path during archive extraction: "
                f"'{finfo.filename}'.",
                stacklevel=2,
            )


def filter_safe_tarinfos(members, base_dir):
    base_dir = resolve_path(base_dir)
    for finfo in members:
        valid_path = False
        if finfo.issym() or finfo.islnk():
            if is_link_in_dir(finfo, base_dir):
                valid_path = True
                yield finfo
        elif resolve_sub_path(base_dir, finfo.name) is not None:
            valid_path = True
            yield finfo
        if not valid_path:
            warnings.warn(
                "Skipping invalid path during archive extraction: "
                f"'{finfo.name}'.",
                stacklevel=2,
            )


def extract_open_archive(archive, path="."):
    """Extracts an open tar or zip archive to the provided directory.

    This function filters unsafe paths during extraction.

    Args:
        archive: The archive object, either a `TarFile` or a `ZipFile`.
        path: Where to extract the archive file.
    """
    if isinstance(archive, zipfile.ZipFile):
        # Zip archive.
        archive.extractall(
            path, members=filter_safe_zipinfos(archive.infolist(), path)
        )
    else:
        # Tar archive.
        extractall_kwargs = {}
        # The `filter="data"` option was:
        # - added in Python 3.12
        # - backported to Python 3.10.12 and 3.11.4
        # - became the default starting from Python 3.14
        # So we only specify it for supported versions before 3.14.
        if (
            sys.version_info < (3, 14)
            and "filter" in inspect.signature(archive.extractall).parameters
        ):
            extractall_kwargs = {"filter": "data"}
        archive.extractall(
            path,
            members=filter_safe_tarinfos(archive, path),
            **extractall_kwargs,
        )


def extract_archive(file_path, path=".", archive_format="auto"):
    """Extracts an archive if it matches a support format.

    Supports `.tar`, `.tar.gz`, `.tar.bz`, and `.zip` formats.

    Args:
        file_path: Path to the archive file.
        path: Where to extract the archive file.
        archive_format: Archive format to try for extracting the file.
            Options are `"auto"`, `"tar"`, `"zip"`, and `None`.
            `"tar"` includes `.tar`, `.tar.gz`, and `.tar.bz` files.
            The default `"auto"` uses `["tar", "zip"]`.
            `None` or an empty list will return no matches found.

    Returns:
        `True` if a match was found and an archive extraction was completed,
        `False` otherwise.
    """
    if archive_format is None:
        return False
    if archive_format == "auto":
        archive_format = ["tar", "zip"]
    if isinstance(archive_format, str):
        archive_format = [archive_format]

    file_path = path_to_string(file_path)
    path = path_to_string(path)

    for archive_type in archive_format:
        if archive_type == "tar":
            open_fn = tarfile.open
            is_match_fn = tarfile.is_tarfile
        elif archive_type == "zip":
            open_fn = zipfile.ZipFile
            is_match_fn = zipfile.is_zipfile
        else:
            raise NotImplementedError(archive_type)

        if is_match_fn(file_path):
            with open_fn(file_path) as archive:
                try:
                    extract_open_archive(archive, path)
                except (tarfile.TarError, RuntimeError, KeyboardInterrupt):
                    if os.path.exists(path):
                        if os.path.isfile(path):
                            os.remove(path)
                        else:
                            shutil.rmtree(path)
                    raise
            return True
    return False


@keras_export("keras.utils.get_file")
def get_file(
    fname=None,
    origin=None,
    untar=False,
    md5_hash=None,
    file_hash=None,
    cache_subdir="datasets",
    hash_algorithm="auto",
    extract=False,
    archive_format="auto",
    cache_dir=None,
    force_download=False,
):
    """Downloads a file from a URL if it not already in the cache.

    By default the file at the url `origin` is downloaded to the
    cache_dir `~/.keras`, placed in the cache_subdir `datasets`,
    and given the filename `fname`. The final location of a file
    `example.txt` would therefore be `~/.keras/datasets/example.txt`.
    Files in `.tar`, `.tar.gz`, `.tar.bz`, and `.zip` formats can
    also be extracted.

    Passing a hash will verify the file after download. The command line
    programs `shasum` and `sha256sum` can compute the hash.

    Example:

    ```python
    path_to_downloaded_file = get_file(
        origin="https://storage.googleapis.com/download.tensorflow.org/example_images/flower_photos.tgz",
        extract=True
    )
    ```

    Args:
        fname: If the target is a single file, this is your desired
            local name for the file.
            If `None`, the name of the file at `origin` will be used.
            If downloading and extracting a directory archive,
            the provided `fname` will be used as extraction directory
            name (only if it doesn't have an extension).
        origin: Original URL of the file.
        untar: Deprecated in favor of `extract` argument.
            Boolean, whether the file is a tar archive that should
            be extracted.
        md5_hash: Deprecated in favor of `file_hash` argument.
            md5 hash of the file for file integrity verification.
        file_hash: The expected hash string of the file after download.
            The sha256 and md5 hash algorithms are both supported.
        cache_subdir: Subdirectory under the Keras cache dir where the file is
            saved. If an absolute path, e.g. `"/path/to/folder"` is
            specified, the file will be saved at that location.
        hash_algorithm: Select the hash algorithm to verify the file.
            options are `"md5'`, `"sha256'`, and `"auto'`.
            The default 'auto' detects the hash algorithm in use.
        extract: If `True`, extracts the archive. Only applicable to compressed
            archive files like tar or zip.
        archive_format: Archive format to try for extracting the file.
            Options are `"auto'`, `"tar'`, `"zip'`, and `None`.
            `"tar"` includes tar, tar.gz, and tar.bz files.
            The default `"auto"` corresponds to `["tar", "zip"]`.
            None or an empty list will return no matches found.
        cache_dir: Location to store cached files, when None it
            defaults ether `$KERAS_HOME` if the `KERAS_HOME` environment
            variable is set or `~/.keras/`.
        force_download: If `True`, the file will always be re-downloaded
            regardless of the cache state.

    Returns:
        Path to the downloaded file.

    **⚠️ Warning on malicious downloads ⚠️**

    Downloading something from the Internet carries a risk.
    NEVER download a file/archive if you do not trust the source.
    We recommend that you specify the `file_hash` argument
    (if the hash of the source file is known) to make sure that the file you
    are getting is the one you expect.
    """
    if origin is None:
        raise ValueError(
            'Please specify the "origin" argument (URL of the file '
            "to download)."
        )

    if cache_dir is None:
        cache_dir = config.keras_home()
    if md5_hash is not None and file_hash is None:
        file_hash = md5_hash
        hash_algorithm = "md5"
    datadir_base = os.path.expanduser(cache_dir)
    if not os.access(datadir_base, os.W_OK):
        datadir_base = os.path.join(
            "/tmp" if os.path.isdir("/tmp") else tempfile.gettempdir(), ".keras"
        )
    datadir = os.path.join(datadir_base, cache_subdir)
    os.makedirs(datadir, exist_ok=True)

    provided_fname = fname
    fname = path_to_string(fname)

    if not fname:
        fname = os.path.basename(urllib.parse.urlsplit(origin).path)
        if not fname:
            raise ValueError(
                "Can't parse the file name from the origin provided: "
                f"'{origin}'."
                "Please specify the `fname` argument."
            )
    else:
        if os.sep in fname:
            raise ValueError(
                "Paths are no longer accepted as the `fname` argument. "
                "To specify the file's parent directory, use "
                f"the `cache_dir` argument. Received: fname={fname}"
            )

    if extract or untar:
        if provided_fname:
            if "." in fname:
                download_target = os.path.join(datadir, fname)
                fname = fname[: fname.find(".")]
                extraction_dir = os.path.join(datadir, f"{fname}_extracted")
            else:
                extraction_dir = os.path.join(datadir, fname)
                download_target = os.path.join(datadir, f"{fname}_archive")
        else:
            extraction_dir = os.path.join(datadir, fname)
            download_target = os.path.join(datadir, f"{fname}_archive")
    else:
        download_target = os.path.join(datadir, fname)

    if force_download:
        download = True
    elif os.path.exists(download_target):
        # File found in cache.
        download = False
        # Verify integrity if a hash was provided.
        if file_hash is not None:
            if not validate_file(
                download_target, file_hash, algorithm=hash_algorithm
            ):
                io_utils.print_msg(
                    "A local file was found, but it seems to be "
                    f"incomplete or outdated because the {hash_algorithm} "
                    "file hash does not match the original value of "
                    f"{file_hash} so we will re-download the data."
                )
                download = True
    else:
        download = True

    if download:
        io_utils.print_msg(f"Downloading data from {origin}")

        class DLProgbar:
            """Manage progress bar state for use in urlretrieve."""

            def __init__(self):
                self.progbar = None
                self.finished = False

            def __call__(self, block_num, block_size, total_size):
                if total_size == -1:
                    total_size = None
                if not self.progbar:
                    self.progbar = Progbar(total_size)
                current = block_num * block_size

                if total_size is None:
                    self.progbar.update(current)
                else:
                    if current < total_size:
                        self.progbar.update(current)
                    elif not self.finished:
                        self.progbar.update(self.progbar.target)
                        self.finished = True

        error_msg = "URL fetch failure on {}: {} -- {}"
        try:
            try:
                urlretrieve(origin, download_target, DLProgbar())
            except urllib.error.HTTPError as e:
                raise Exception(error_msg.format(origin, e.code, e.msg))
            except urllib.error.URLError as e:
                raise Exception(error_msg.format(origin, e.errno, e.reason))
        except (Exception, KeyboardInterrupt):
            if os.path.exists(download_target):
                os.remove(download_target)
            raise

        # Validate download if succeeded and user provided an expected hash
        # Security conscious users would get the hash of the file from a
        # separate channel and pass it to this API to prevent MITM / corruption:
        if os.path.exists(download_target) and file_hash is not None:
            if not validate_file(
                download_target, file_hash, algorithm=hash_algorithm
            ):
                raise ValueError(
                    "Incomplete or corrupted file detected. "
                    f"The {hash_algorithm} "
                    "file hash does not match the provided value "
                    f"of {file_hash}."
                )

    if extract or untar:
        if untar:
            archive_format = "tar"

        status = extract_archive(
            download_target, extraction_dir, archive_format
        )
        if not status:
            warnings.warn("Could not extract archive.", stacklevel=2)
        return extraction_dir

    return download_target


def resolve_hasher(algorithm, file_hash=None):
    """Returns hash algorithm as hashlib function."""
    if algorithm == "sha256":
        return hashlib.sha256()

    if algorithm == "auto" and file_hash is not None and len(file_hash) == 64:
        return hashlib.sha256()

    # This is used only for legacy purposes.
    return hashlib.md5()


def hash_file(fpath, algorithm="sha256", chunk_size=65535):
    """Calculates a file sha256 or md5 hash.

    Example:

    >>> hash_file('/path/to/file.zip')
    'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'

    Args:
        fpath: Path to the file being validated.
        algorithm: Hash algorithm, one of `"auto"`, `"sha256"`, or `"md5"`.
            The default `"auto"` detects the hash algorithm in use.
        chunk_size: Bytes to read at a time, important for large files.

    Returns:
        The file hash.
    """
    if isinstance(algorithm, str):
        hasher = resolve_hasher(algorithm)
    else:
        hasher = algorithm

    with open(fpath, "rb") as fpath_file:
        for chunk in iter(lambda: fpath_file.read(chunk_size), b""):
            hasher.update(chunk)

    return hasher.hexdigest()


def validate_file(fpath, file_hash, algorithm="auto", chunk_size=65535):
    """Validates a file against a sha256 or md5 hash.

    Args:
        fpath: path to the file being validated
        file_hash:  The expected hash string of the file.
            The sha256 and md5 hash algorithms are both supported.
        algorithm: Hash algorithm, one of `"auto"`, `"sha256"`, or `"md5"`.
            The default `"auto"` detects the hash algorithm in use.
        chunk_size: Bytes to read at a time, important for large files.

    Returns:
        Boolean, whether the file is valid.
    """
    hasher = resolve_hasher(algorithm, file_hash)

    if str(hash_file(fpath, hasher, chunk_size)) == str(file_hash):
        return True
    else:
        return False


def is_remote_path(filepath):
    """
    Determines if a given filepath indicates a remote location.

    This function checks if the filepath represents a known remote pattern
    such as GCS (`/gcs`), CNS (`/cns`), CFS (`/cfs`), HDFS (`/hdfs`), Placer
    (`/placer`), TFHub (`/tfhub`), or a URL (`.*://`).

    Args:
        filepath (str): The path to be checked.

    Returns:
        bool: True if the filepath is a recognized remote path, otherwise False
    """
    if re.match(
        r"^(/cns|/cfs|/gcs|/hdfs|/readahead|/placer|/tfhub|.*://).*$",
        str(filepath),
    ):
        return True
    return False


# Below are gfile-replacement utils.


def _raise_if_no_gfile(path):
    raise ValueError(
        "Handling remote paths requires installing TensorFlow "
        f"(in order to use gfile). Received path: {path}"
    )


def exists(path):
    if is_remote_path(path):
        if gfile.available:
            return gfile.exists(path)
        else:
            _raise_if_no_gfile(path)
    return os.path.exists(path)


def File(path, mode="r"):
    if is_remote_path(path):
        if gfile.available:
            return gfile.GFile(path, mode=mode)
        else:
            _raise_if_no_gfile(path)
    return open(path, mode=mode)


def join(path, *paths):
    if is_remote_path(path):
        if gfile.available:
            return gfile.join(path, *paths)
        else:
            _raise_if_no_gfile(path)
    return os.path.join(path, *paths)


def isdir(path):
    if is_remote_path(path):
        if gfile.available:
            return gfile.isdir(path)
        else:
            _raise_if_no_gfile(path)
    return os.path.isdir(path)


def remove(path):
    if is_remote_path(path):
        if gfile.available:
            return gfile.remove(path)
        else:
            _raise_if_no_gfile(path)
    return os.remove(path)


def rmtree(path):
    if is_remote_path(path):
        if gfile.available:
            return gfile.rmtree(path)
        else:
            _raise_if_no_gfile(path)
    return shutil.rmtree(path)


def listdir(path):
    if is_remote_path(path):
        if gfile.available:
            return gfile.listdir(path)
        else:
            _raise_if_no_gfile(path)
    return os.listdir(path)


def copy(src, dst):
    if is_remote_path(src) or is_remote_path(dst):
        if gfile.available:
            return gfile.copy(src, dst, overwrite=True)
        else:
            _raise_if_no_gfile(f"src={src} dst={dst}")
    return shutil.copy(src, dst)


def makedirs(path):
    if is_remote_path(path):
        if gfile.available:
            return gfile.makedirs(path)
        else:
            _raise_if_no_gfile(path)
    return os.makedirs(path)



import re
from collections.abc import MutableMapping

from keras.src.api_export import keras_export
from keras.src.optimizers.loss_scale_optimizer import LossScaleOptimizer
from keras.src.optimizers.optimizer import Optimizer
from keras.src.saving import serialization_lib


@keras_export("keras.optimizers.OptimizerMap")
class OptimizerMap(MutableMapping):
    """A class mapping variables to optimizers.

    Args:
        default_optimizer: Default Keras `Optimizer`
            for any unmapped variables.
    """

    def __init__(self, default_optimizer, optimizer_map=None):
        if not isinstance(default_optimizer, Optimizer):
            raise TypeError(
                "default_optimizer must be a Keras Optimizer instance. "
                f"Received: {default_optimizer}"
            )
        if optimizer_map is not None and not isinstance(optimizer_map, dict):
            raise TypeError(
                f"optimizer_map must be a dictionary. Received: {optimizer_map}"
            )
        self._default_optimizer = default_optimizer
        self._optimizer_map = optimizer_map or dict()

    @property
    def default_optimizer(self):
        return self._default_optimizer

    def __getitem__(self, key):
        """Retrieves the corresponding `Optimizer` by the string key.

        This method first attempts an exact key match. If no exact match is
        found, it treats all keys in the map as regular expression patterns
        and uses `re.fullmatch` to find a policy.

        For example,
        to apply a optimizer to all sublayers of an `encoder` block,
        the key should be explicitly set to `"encoder/.*"`. A key of
        `"encoder"` will only match the layer with that exact path.

        Args:
            key: str. The key to query for an `Optimizer`.

        Returns:
            The corresponding `Optimizer`. If no match is found, this method
            returns `self.default_optimizer`.

        Raises:
            ValueError: If the `key` matches more than one regex pattern in the
            map.

        Example:

        ```python
        >>> from keras.src import optimizers
        >>> from keras.src.optimizers.multi_optimizer import OptimizerMap
        >>> opt_adam = optimizers.Adam()
        >>> opt_sgd = optimizers.SGD()
        >>> opt_rmsprop = optimizers.RMSprop()
        >>> optimizer_map = OptimizerMap(default_optimizer=opt_rmsprop)
        >>> optimizer_map["encoder/layer_0/dense"] = opt_adam
        >>> optimizer_map["encoder/.*"] = opt_sgd
        >>> optimizer_map["decoder"] = opt_adam

        >>> optimizer_map["decoder"].name
        'adam'

        >>> optimizer_map["encoder/layer_0/dense"].name
        'adam'

        >>> optimizer_map["encoder/layer_0/attention/query"].name
        'sgd'

        >>> optimizer_map["decoder/layer_0/dense"].name
        'rmsprop'
        ```
        """

        if key in self._optimizer_map:
            return self._optimizer_map[key]

        matching_keys = [
            pattern
            for pattern in self._optimizer_map.keys()
            if re.fullmatch(pattern, key)
        ]

        if len(matching_keys) > 1:
            raise ValueError(f"Multiple optimizers assigned to variable {key}.")
        elif len(matching_keys) == 1:
            return self._optimizer_map[matching_keys[0]]
        return self.default_optimizer

    def __setitem__(self, key, optimizer):
        """Set optimizer for a given variable.

        Args:
            key: string representing the variable path or regex pattern
            optimizer: Keras Optimizer instance
        """
        if key in self._optimizer_map:
            raise ValueError(
                f"'{key}' already exists in the OptimizerMap with "
                f"value {self._optimizer_map[key]}. Please make sure to "
                "not use duplicated keys."
            )
        if not isinstance(key, str):
            raise ValueError(f"key must be a string. Received: {key}")
        if not isinstance(optimizer, Optimizer):
            raise ValueError(
                f"optimizer must be a Keras Optimizer instance. "
                f"Received: {optimizer}"
            )
        if isinstance(optimizer, LossScaleOptimizer):
            raise ValueError("optimizer cannot be LossScaleOptimizer.")
        if isinstance(optimizer, MultiOptimizer):
            raise ValueError("optimizer cannot be MultiOptimizer.")
        self._optimizer_map[key] = optimizer

    def __delitem__(self, key):
        del self._optimizer_map[key]

    def __len__(self):
        return len(self._optimizer_map)

    def __iter__(self):
        return iter(self._optimizer_map)

    def __call__(self, variable):
        return self[variable.path]

    def get_config(self):
        return {
            "default_optimizer": serialization_lib.serialize_keras_object(
                self._default_optimizer
            ),
            "optimizer_map": serialization_lib.serialize_keras_object(
                self._optimizer_map
            ),
        }

    @classmethod
    def from_config(cls, config, custom_objects=None):
        default_optimizer = serialization_lib.deserialize_keras_object(
            config["default_optimizer"], custom_objects=custom_objects
        )
        optimizer_map = serialization_lib.deserialize_keras_object(
            config["optimizer_map"], custom_objects=custom_objects
        )
        obj_map = cls(default_optimizer, optimizer_map)
        return obj_map


@keras_export("keras.optimizers.MultiOptimizer")
class MultiOptimizer(Optimizer):
    """An optimizer wrapper that delegates variables to different optimizers.

    Initialize the object with an OptimizerMap instance or a callable
    function that returns an optimizer for a given variable.

    Example:
        model.compile(
            optimizer=MultiOptimizer(
                OptimizerMap(default_optimizer=optimizers.SGD(),
                {"encoder/.*": optimizers.Adam()})
            ),
            loss="binary_crossentropy",
        )

        # Or using a callable
        def optimizer_selector(variable):
            if "encoder" in variable.path:
                return optimizers.Adam()
            else:
                return optimizers.SGD()

        model.compile(
            optimizer=MultiOptimizer(optimizer_selector),
            loss="binary_crossentropy",
        )

    To access the attributes of the sub-optimizers, iterate over the
    optimizers using `.optimizers`:

    For example:

    optimizer = MultiOptimizer(OptimizerMap(
        default_optimizer=optimizers.Adam()
    ))
    optimizer['.encoder'] = optimizers.SGD()

    for optim in optimizer.optimizers:
        print(optim.learning_rate)
        print(optim.iterations)
        print(optim.loss_scale_factor)
        ...

    The MultiOptimizer class instances will not expose `learning_rate`
    attribute and will raise an error if accessed. This is because the
    learning rate might be different for different sub-optimizers.

    Note: Optimizer-specific callbacks are not supported yet.

    """

    def __init__(self, optimizer_map, loss_scale_factor=None, name=None):
        """
        Initialize the MultiOptimizer.

        Args:
            optimizer_map: An OptimizerMap instance or a callable function that
                returns an optimizer for a given variable.
            loss_scale_factor: It overrides the loss_scale_factor passed
            to the sub-optimizers.
            name: The name of the optimizer.
        """
        super().__init__(
            learning_rate=0.0,
            loss_scale_factor=loss_scale_factor,
            name=name,
        )
        self._optimizer_map = optimizer_map
        self._inner_optimizers = []

        if hasattr(self._optimizer_map, "values"):
            for opt in self._optimizer_map.values():
                if opt not in self._inner_optimizers:
                    opt.loss_scale_factor = loss_scale_factor
                    self._inner_optimizers.append(opt)

        default_opt = getattr(self._optimizer_map, "default_optimizer", None)
        if (
            default_opt is not None
            and default_opt not in self._inner_optimizers
        ):
            default_opt.loss_scale_factor = loss_scale_factor
            self._inner_optimizers.append(default_opt)

    def build(self, var_list):
        if self.built:
            return

        self._var_to_optimizer_idx = {}
        self._trainable_variables_indices = {}
        optimizer_vars = [[] for _ in range(len(self._inner_optimizers))]

        for i, var in enumerate(var_list):
            opt = self._optimizer_map(var)
            if not isinstance(opt, Optimizer):
                raise ValueError(
                    f"Optimizer for variable {var} is not "
                    "an Optimizer instance."
                )
            if isinstance(opt, LossScaleOptimizer):
                raise ValueError(
                    "LossScaleOptimizer cannot be used inside an "
                    "MultiOptimizer."
                )
            if isinstance(opt, MultiOptimizer):
                raise ValueError(
                    "MultiOptimizer cannot be used inside an MultiOptimizer."
                )
            if opt not in self._inner_optimizers:
                opt.loss_scale_factor = self.loss_scale_factor
                self._inner_optimizers.append(opt)
                optimizer_vars.append([])

            idx = self._inner_optimizers.index(opt)
            self._var_to_optimizer_idx[self._var_key(var)] = idx
            self._trainable_variables_indices[self._var_key(var)] = i
            optimizer_vars[idx].append(var)

        # Build all sub-optimizers
        for opt, variables in zip(self._inner_optimizers, optimizer_vars):
            if variables:
                opt.build(variables)

        super().build(var_list)

    @property
    def optimizers(self):
        return self._inner_optimizers

    @property
    def variables(self):
        # MultiOptimizer iterations + all sub-optimizer variables
        vars = self._variables[:]
        for opt in self._inner_optimizers:
            vars.extend(opt.variables)
        return vars

    @property
    def learning_rate(self):
        raise AttributeError(
            "Learning rate cannot be accessed on a MultiOptimizer. "
            "Access the learning rate on the individual sub-optimizers instead."
        )

    @learning_rate.setter
    def learning_rate(self, value):
        raise AttributeError(
            "Learning rate cannot be set on a MultiOptimizer. "
            "Set the learning rate on the individual sub-optimizers instead."
        )

    @property
    def loss_scale_factor(self):
        return getattr(self, "_loss_scale_factor", None)

    @loss_scale_factor.setter
    def loss_scale_factor(self, value):
        self._loss_scale_factor = value
        if hasattr(self, "_inner_optimizers"):
            for opt in self._inner_optimizers:
                opt.loss_scale_factor = value

    @property
    def iterations(self):
        return self._iterations

    def apply(self, grads, trainable_variables=None):
        if len(grads) == 0:
            return

        if trainable_variables is None:
            if not self.built:
                raise ValueError(
                    "When passing `grads` without `variables`, the optimizer "
                    "must already be built on a list of variables. "
                    "Call `optimizer.build(trainable_variables)` first."
                )
            trainable_variables = self._trainable_variables

        if not self.built:
            self.build(trainable_variables)
            self.built = True

        if len(grads) != len(self._trainable_variables_indices):
            raise ValueError(
                "Gradients must match trainable variables one-to-one. "
                f"Received {len(grads)} gradients and "
                f"{len(self._trainable_variables_indices)} variables."
            )

        grads_and_vars = [[] for _ in range(len(self._inner_optimizers))]
        for grad, var in zip(grads, trainable_variables):
            idx = self._var_to_optimizer_idx[self._var_key(var)]
            grads_and_vars[idx].append((grad, var))

        # Dispatch apply calls
        for opt, sub_grads_and_vars in zip(
            self._inner_optimizers, grads_and_vars
        ):
            if sub_grads_and_vars:
                sub_grads, sub_vars = zip(*sub_grads_and_vars)
                opt.apply(list(sub_grads), list(sub_vars))

        self._iterations.assign_add(1)

    def finalize_variable_values(self, var_list):
        optimizer_vars = [[] for _ in range(len(self._inner_optimizers))]
        for var in var_list:
            idx = self._var_to_optimizer_idx[self._var_key(var)]
            optimizer_vars[idx].append(var)

        for opt, variables in zip(self._inner_optimizers, optimizer_vars):
            if variables:
                opt.finalize_variable_values(variables)

    def set_weights(self, weights):
        if not self.built:
            raise ValueError(
                "You are calling `set_weights()` on an optimizer that has not "
                "yet been built. Please call "
                "`optimizer.build(trainable_variables)` to create the "
                "optimizer weights before calling `set_weights()`."
            )

        own_var_count = len(self._variables)
        for i in range(own_var_count):
            self._variables[i].assign(weights[i])

        idx = own_var_count
        for opt in self._inner_optimizers:
            num_opt_vars = len(opt.variables)
            if num_opt_vars > 0:
                opt_weights = weights[idx : idx + num_opt_vars]
                opt.set_weights(opt_weights)
                idx += num_opt_vars

    def get_config(self):
        return {
            "optimizer_map": serialization_lib.serialize_keras_object(
                self._optimizer_map
            ),
            "loss_scale_factor": self.loss_scale_factor,
            "name": self.name,
        }

    @classmethod
    def from_config(cls, config, custom_objects=None):
        config = config.copy()
        config["optimizer_map"] = serialization_lib.deserialize_keras_object(
            config["optimizer_map"], custom_objects=custom_objects
        )
        return cls(**config)
